'''
CNA_Database.py, Geoffrey Weal, 29/10/2018

This script holds the information required to make a CNA_database.

'''
import os, sys, time
import multiprocessing as mp
from Organisms.GA.SCM_Scripts.MyPool import MyPool
from Organisms.GA.SCM_Scripts.Similarity_Profile import Similarity_Profile

class Tree(dict): #WeakValueDictionary()
	"""
	This is a Tree designed for the CNA_Database to hold references of CNA_Entry.
	"""
	def __missing__(self, key):
		"""
		This rewrites the __missing__ component of a dict to work as a tree.
		"""
		value = self[key] = type(self)()
		return value

	def see_tree(self):
		"""
		This shown the clusters in the database
		"""
		toString = ''
		for key, value in self:
			toString += str(key)+': '+str(sorted([key2 for key2, value2 in self[key]]))+'\n'
		return toString

def cna_profile_generator(collection,rCuts):
	"""
	This is a generator that returns the clusters in the collection with the rCut values to scan across. 

	:param collection: A collection
	:type  collection: Organisms.GA.Collection
	:param rCuts: The list of clusters to scan across with the SCM. 
	:type  rCuts: list of float

	:returns: a tuple of the cluster the rCut values to scan across with the SCM
	"""
	for cluster in collection:
		yield cluster,rCuts

def initial_similarity_profile_generator(collection,cna_database):
	"""
	This is a generator that returns the clusters in the collection with the rCut values to scan across. 

	:param collection: A collection
	:type  collection: Organisms.GA.Collection
	:param cna_database: The list of clusters to scan across with the SCM. 
	:type  cna_database: list of float

	:returns: a tuple of the names of the clusters and their associated CNA profiles.
	:rtype:   (int, int, Counter, Counter)
	"""
	for index_1 in range(len(collection)):
		name_1 = collection[index_1].name
		for index_2 in range(index_1+1,len(collection)):
			name_2 = collection[index_2].name
			if not cna_database.is_pair_in_the_database(name_1,name_2):
				cna_profile_1 = cna_database.cna_profile_database[name_1]
				cna_profile_2 = cna_database.cna_profile_database[name_2]
				yield name_1, name_2, cna_profile_1, cna_profile_2

def similarity_profile_generator(population,offsprings,cna_database):
	"""
	This is a generator that returns the clusters in the collection with the rCut values to scan across. 

	:param population: The population
	:type  population: Organisms.GA.Population
	:param offsprings: The collection of offspring
	:type  offsprings: Organisms.GA.Offspring_Pool
	:param cna_database: The list of clusters to scan across with the SCM. 
	:type  cna_database: list of float

	:returns: a tuple of the names of the clusters and their associated CNA profiles.
	:rtype:   (int, int, Counter, Counter)
	"""
	for cluster_pop in population:
		for offspring in offsprings:
			name_pop = cluster_pop.name
			name_off = offspring.name
			if not cna_database.is_pair_in_the_database(name_pop,name_off):
				cna_profile_cluster_pop = cna_database.cna_profile_database[name_pop]
				cna_profile_offspring   = cna_database.cna_profile_database[name_off]
				yield name_pop, name_off, cna_profile_cluster_pop, cna_profile_offspring
	#initial_similarity_profile_generator(offsprings,cna_database)
	for index_1 in range(len(offsprings)):
		name_1 = offsprings[index_1].name
		for index_2 in range(index_1+1,len(offsprings)):
			name_2 = offsprings[index_2].name
			if not cna_database.is_pair_in_the_database(name_1,name_2):
				cna_profile_1 = cna_database.cna_profile_database[name_1]
				cna_profile_2 = cna_database.cna_profile_database[name_2]
				yield name_1, name_2, cna_profile_1, cna_profile_2

class CNA_Database:
	"""
	This is a database that holds all the entries of CNA_Entry objects.

	:param rCuts: These are the rCut values to scan the CNA across. 
	:type  rCuts: float
	:param population: 
	:type  population: Organisms.GA.Population
	:param cut_off_similarity: The maximum similarity above which to be considered similar enough to exclude offspring from being accessed into future generations. 
	:type  cut_off_similarity: float
	:param get_cna_profile_method: This is the method to get the CNA Profile, either from the T-SCM or the A-SCM. 
	:type  get_cna_profile_method: __func__
	:param get_similarity_profile_method: This is the method that uses the CNA profiles of two clusters in order to give the similarity profile between the two clusters, Whether this is obtained by the T-SCM or the A-SCM. 
	:type  get_similarity_profile_method: __func__
	:param no_of_cpus: The number of cpu to use to obtain the similarity profile between two clusters
	:type  no_of_cpus: int
	:param debug: Get data to use to debug the SCM. Default: False
	:type  debug: bool. 

	"""
	def __init__(self,rCuts,population,cut_off_similarity,get_cna_profile_method,get_similarity_profile_method,no_of_cpus,debug):

		self.cna_profile_database = {}
		self.similarity_profile_database = Tree()

		self.rCuts = rCuts
		self.population = population
		self.cut_off_similarity = cut_off_similarity
		self.get_cna_profile_method = get_cna_profile_method
		self.get_similarity_profile_method = get_similarity_profile_method
		self.no_of_cpus = no_of_cpus
		self.debug = debug

	def reset(self):
		"""
		Reset the CNA profiles of generated clusters and the similarity profile database.
		"""
		self.cna_profile_database = {}
		self.similarity_profile_database = Tree()

	def get_details(self):
		"""
		Retun the information about the CNA Database, including:

		:returns: The cut_off_similarity, get_cna_profile_method, get_similarity_profile_method, no_of_cpus, debug
		:rtype:   (float, __func__, __func__, int, bool)
		"""
		return (self.cut_off_similarity, self.get_cna_profile_method, self.get_similarity_profile_method, self.no_of_cpus, self.debug)

	def __len__(self):
		"""
		Get the number of clusters in the database.

		:returns: The number of CNA profiles that have been recorded. 
		:rtype:   int
		"""
		return len(self.cna_profile_database)

	def keys(self):
		"""
		Return a list of the names of all the clusters in the database

		:returns: A list of the names of all the clusters in the database
		:rtype:   list of int
		"""
		return list(self.cna_profile_database.keys())

	def __getitem__(self,name):
		"""
		Return the similariy_profile of cluster name:

		:param name: The name of the cluster
		:type  name: int

		:returns: The similarity profile for cluster name.
		:rtype:   list of float
		"""
		try:
			return self.similarity_profile_database[name]
		except:
			print("Error in Def __getitem__ in class CNA_Database, in CNA_Database.py")
			print('You are trying to get cluster '+str(name))
			print('But this does not exist in the CNA database')
			print('Clusters in the database: '+str(self.keys()))
			print('Check this out')
			import pdb; pdb.set_trace()
			exit('This program will exit without completing')

	def add(self,collection,initialise=False):
		"""
		Add the similarity data of the clusters in the collection to the CNA_Database

		:param collection: The clusters to add to the CNA_Database
		:type  collection: Organisms.GA.Collection
		:param initialise: Are the clusters from a newly created population. Default: False
		:type  initialise: bool
		"""
		# First, add all the CNA profiles to the cna_profile_database
		CNA_Profiles_to_make = cna_profile_generator(collection,self.rCuts)
		start_time = time.time()
		with MyPool(processes=self.no_of_cpus) as pool:
			results = pool.map_async(self.get_cna_profile_method, CNA_Profiles_to_make)
			results.wait()
		data = results.get()
		end_time = time.time()
		print('Processing CNA Profiles took '+str(end_time - start_time)+' seconds.')
		for cluster_dir, cluster_CNA_profile in data:
			self.cna_profile_database[cluster_dir] = cluster_CNA_profile
		# Second, get the the similarity profiles and add them to the database.
		if initialise:
			CNA_Similarities_to_make = initial_similarity_profile_generator(collection,self)
		else:
			CNA_Similarities_to_make = similarity_profile_generator(self.population,collection,self)
		start_time = time.time()
		with mp.Pool(processes=self.no_of_cpus) as pool:
			results = pool.map_async(self.get_similarity_profile_method, CNA_Similarities_to_make)
			results.wait()
		data = results.get()
		end_time = time.time()
		print('Processing CNA similarities took '+str(end_time - start_time)+' seconds.')
		for name_1, name_2, cna_similarity in data:
			similarity_profile = Similarity_Profile(name_1, name_2, cna_similarity)
			self.similarity_profile_database[name_1][name_2] = similarity_profile
			self.similarity_profile_database[name_2][name_1] = similarity_profile
		##########################################################################

	def remove(self,names_to_remove):
		"""
		Remove all entries that exists in the database that are associated with a particular cluster.

		:param name_to_remove: The name of the cluster you would like to remove from the CNA_Database.
		:type  name_to_remove: int

		"""
		# Take a note of the clusters that will not be removed from the CNA_Database
		# These will be used to help efficiently remove cluster entries in similarity_profile_database
		# that are not needed anymore.
		names_to_look_though = list(self.similarity_profile_database.keys())
		for name_to_remove in names_to_remove:
			if name_to_remove in names_to_look_though:
				names_to_look_though.remove(name_to_remove)
		# remove all cluster entries that are no longer needed in self.cna_profile_database and self.similarity_profile_database
		for name_to_remove in names_to_remove:
			del self.cna_profile_database[name_to_remove]
			del self.similarity_profile_database[name_to_remove]
			for name_to_look_at in names_to_look_though:
				del self.similarity_profile_database[name_to_look_at][name_to_remove]

	def is_pair_in_the_database(self,dir_1,dir_2):
		"""
		Determine if a CNA entry exists for two clusters in the similarity_profile_database

		:param name1: The dir of the first cluster you would like to look for an entry in the CNA_Database.
		:type  name1: int
		:param name2: The dir of the second cluster you would like to look for an entry in the CNA_Database.
		:type  name2: int

		:returns: True if the similarity profile for cluster name1 and name2, False if not.
		:rtype:   bool
		"""
		if dir_1 == dir_2:
			print('Error in class CNA_Database, in CNA_Database.py')
			print('dir_1 and dir_2 are the same')
			print('dir_1 = '+str(dir_1))
			print('dir_2 = '+str(dir_2))
			print('Check this.')
			import pdb; pdb.set_trace()
			exit('This program will exit without completing')
		if self.similarity_profile_database[dir_1][dir_2] == {}:
			if not self.similarity_profile_database[dir_2][dir_1] == {}:
				print('Error in class CNA_Database, in CNA_Database.py')
				print('The Tree is not full')
				print('There is an entry for self.similarity_profile_database[dir_1][dir_2], but not for self.similarity_profile_database[dir_2][dir_1].')
				print('dir_1 = '+str(dir_1))
				print('dir_2 = '+str(dir_2))
				print('self.similarity_profile_database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.similarity_profile_database[dir_1][dir_2]))
				print('self.similarity_profile_database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.similarity_profile_database[dir_2][dir_1]))
				print('This should not be the case. What should be true is self.similarity_profile_database[dir_1][dir_2] == self.similarity_profile_database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
				print('Check this.')
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			del self.similarity_profile_database[dir_2][dir_1]
			del self.similarity_profile_database[dir_1][dir_2]
			return False
		else:
			if self.similarity_profile_database[dir_2][dir_1] == {}:
				print('Error in class CNA_Database, in CNA_Database.py')
				print('The Tree is not full')
				print('There is an entry for self.similarity_profile_database[dir_1][dir_2], but not for self.similarity_profile_database[dir_2][dir_1].')
				print('dir_1 = '+str(dir_1))
				print('dir_2 = '+str(dir_2))
				print('self.similarity_profile_database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.similarity_profile_database[dir_1][dir_2]))
				print('self.similarity_profile_database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.similarity_profile_database[dir_2][dir_1]))
				print('This should not be the case. What should be true is self.similarity_profile_database[dir_1][dir_2] == self.similarity_profile_database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
				print('Check this.')
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			elif not self.similarity_profile_database[dir_1][dir_2] is self.similarity_profile_database[dir_2][dir_1]:
				print('Error in class CNA_Database, in CNA_Database.py')
				print('self.similarity_profile_database[dir_1][dir_2] is not the same as self.similarity_profile_database[dir_2][dir_1]')
				print('dir_1 = '+str(dir_1))
				print('dir_2 = '+str(dir_2))
				print('self.similarity_profile_database['+str(dir_1)+']['+str(dir_2)+'] = '+str(self.similarity_profile_database[dir_1][dir_2]))
				print('self.similarity_profile_database['+str(dir_2)+']['+str(dir_1)+'] = '+str(self.similarity_profile_database[dir_2][dir_1]))
				print('This should not be the case. What should be true is self.similarity_profile_database[dir_1][dir_2] == self.similarity_profile_database[dir_2][dir_1] == CNA_Entry(cluster1,cluster2).')
				print('Check this.')
				import pdb; pdb.set_trace()
				exit('This program will exit without completing')
			return True

	def get_similar_clusters_in_database(self):
		"""
		Get clusters in the clusters that are deemed structurally similar in the CNA_Database

		:returns: a list of the names of all the clusters thaat are similar to each other.
		:rtype:   list of int
		"""
		too_similar_database = {}
		names_in_database = list(self.keys())
		for index1 in range(len(names_in_database)):
			name1 = names_in_database[index1]
			for index2 in range(index1+1,len(names_in_database)):
				name2 = names_in_database[index2]
				if self.get_max_similarity(name1,name2) >= self.cut_off_similarity:
					too_similar_database.setdefault(name1,[]).append(name2)
					too_similar_database.setdefault(name2,[]).append(name1)
		return too_similar_database

	def get_all_averages_for_a_cluster(self,cluster_name):
		"""
		Get the averages SCM similarities of a cluser compared to every other clusrer in the similarity profile.

		:param cluster_name: Cluster to get the averages similarities of.
		:type  cluster_name: list of floats

		:returns: all the average similarities between cluster_name any every other cluster in the database. 
		:rtype:   list of float
		"""
		all_cna_averages = [cna_entry.average for _,cna_entry in self.similarity_profile_database[cluster_name].items()]
		return all_cna_averages

	def check_database(self, population):
		"""
		Check the database to make sure there are the correct number of entries in the database.

		:param population: The population
		:type  population: Organisms.GA.Population
		"""
		if not len(self.similarity_profile_database) == len(population):
			print('Issue') 
			import pdb; pdb.set_trace()
			exit('Issue') 
		for branch in self.similarity_profile_database.values():
			if not len(branch) == len(population)-1:
				print('Issue') 
				import pdb; pdb.set_trace()
				exit('Issue') 

	def get_max_similarity(self,name_1,name_2):
		"""
		This def will obtain the max similarity percentage from the compararison of two clusters, name_1 and name_2.

		:param name_1: The name of the first cluster you would like to look for an entry in the CNA_Database.
		:type  name_1: int
		:param name_2: The name of the second cluster you would like to look for an entry in the CNA_Database.
		:type  name_2: int

		:returns: returns the maximum similarity between cluster name_1 and name_2
		:rtype:   float
		"""
		similarity_profile = self.similarity_profile_database[name_1][name_2]
		try:
			return similarity_profile.similarity_max
		except:
			print('Error in def CNA_Analysis of Class CNA_Database of Diversity Scheme.')
			print('This def can not obtain a similarity_category')
			print('This probably means that an entry does not exist')
			print('name_1 = '+str(name_1))
			print('name_2 = '+str(name_2))
			print('self.database['+str(name_1)+']['+str(name_2)+'] = '+str(similarity_profile))
			print('Other information')
			print('self.similarity_profile_database['+str(name_1)+'] = '+str(self.similarity_profile_database[name_1]))
			print('self.similarity_profile_database['+str(name_2)+'] = '+str(self.similarity_profile_database[name_2]))
			#self.print_cna_database_details()
			print('Check this out.')
			import pdb; pdb.set_trace()
			exit()

	def print_cna_database_details(self):
		"""
		Print information about the database. 
		"""
		print('Clusters recorded in the CNA Profile Database: '+str(list(self.cna_profile_database.keys())))
		print('Clusters recorded in the Similarity Profile Database: '+str(list(self.similarity_profile_database.keys())))
		print('')
		print('Average similarities between clusters in the Similarity Profile Database:')
		self.make_simple_table(self.similarity_profile_database)

	def make_simple_table(self,similarity_profile_database):
		"""
		This is a simpled table that can be printed to the terminal that shows all the similarities between clusters in the population + offspring

		:param similarity_profile_database: This is the CNA database that contains all similarity information of clusters in rhe population and offspring.
		:type  similarity_profile_database: Organisms.GA.SCM_Scripts.CNA_Database.CNA_Database

		"""
		similarity_profile_database_keys = sorted(list(similarity_profile_database.keys()))
		similarity_profile_database_list = sorted(list(similarity_profile_database.items()))
		row_format ="{:>4}" * (len(similarity_profile_database_list) + 1)
		print(row_format.format("", *similarity_profile_database_keys))
		for cluster_name, similarity_data in similarity_profile_database_list:
			similarity_data = list(similarity_data.items())
			similarity_data_format = []
			for column, similarity_datum in similarity_data:
				try:
					similarity_data_format.append((column, str(similarity_datum.average)))
				except:
					similarity_data_format.append((column, str(similarity_datum)))
			similarity_data_format += [(cluster_name,'-')]
			similarity_data_format.sort()
			import pdb; pdb.set_trace()
			for index in range(len(similarity_data_format)):
				similarity_data_format[index] = similarity_data_format[index][1]
			import pdb; pdb.set_trace()
			print(row_format.format(cluster_name, *similarity_data_format))











def mean(all_cna_averages):
	return (sum(all_cna_averages) / len(all_cna_averages))

def get_CNA_most_similar_average(cluster,cna_database,collection_function):
	cluster_name = cluster.name
	all_cna_averages = cna_database.get_all_averages_for_a_cluster(cluster_name)
	CNA_most_similar_average = float(collection_function(all_cna_averages))
	return CNA_most_similar_average

def get_CNA_fitness_parameter(cluster,cna_database,collection_function):
	"""
	Get the order parameter, based on the CNA, for a cluster. 

	:param cluster: This is the cluster you want to obtain rho_i for.
	:type  cluster: Organisms.GA.Cluster
	:param cna_database: This is an in-memory database (not a ASE disk database) that records the similarities between clusters in the population.
	:type  cna_database: Organisms.GA.SCM_Scripts.CNA_Database
	:param cna_fitness_function: This is the function that converts the rho similarity into a fitness value. 
	:type  cna_fitness_function: __func__

	:returns: CNA_fitness_value: This is the SRM fitness contribution to be used to obtain the fitness valuee 
	:rtypes: float

	"""
	CNA_most_similar_average = get_CNA_most_similar_average(cluster,cna_database,collection_function)
	# Change this to an fitness parameter, where: # Changed 3/12/2018, GRW
	# 	* High similarity -> large fitness parameter -> less favourable for mating and natural selection
	# 	* Low similarity  -> small fitness parameter -> more favourable for mating and natural selection
	CNA_fitness_contribution = (CNA_most_similar_average/100.0) # Note from GRW. This term should be CNA_most_similar_average/100.0
	return CNA_fitness_contribution

def get_CNA_fitness_contribution(cluster,max_similarity,min_minimarity,cna_database,collection_function,cna_fitness_function):
	"""
	Get the fitness, based on the structural diversity as determined by the SCM, for a cluster. 

	:param cluster: This is the cluster you want to obtain rho_i for.
	:type  cluster: Organisms.GA.Cluster
	:param max_similarity: The maximum similarity obtained from the population + offspring_pools. THIS IS NOT USED IN THIS METHOD, IS HERE FOR CONSISTANCY.
	:type  max_similarity: Not needed to be anything
	:param min_minimarity: The minimum similarity obtained from the population + offspring_pools. THIS IS NOT USED IN THIS METHOD, IS HERE FOR CONSISTANCY.
	:type  min_minimarity: Not needed to be anything
	:param cna_database: This is an in-memory database (not a ASE disk database) that records the similarities between clusters in the population.
	:type  cna_database: Organisms.GA.SCM_Scripts.CNA_Database
	:param cna_fitness_function: This is the function that converts the rho similarity into a fitness value. 
	:type  cna_fitness_function: __func__

	"""
	rho_i = get_CNA_fitness_parameter(cluster,cna_database,collection_function)
	fitness = cna_fitness_function.get_fitness(rho_i)
	return fitness, rho_i

# ------------------------------------------------------------------------------------------------------------------------------------------------

def get_CNA_fitness_parameter_normalised(cluster,max_similarity,min_minimarity,cna_database,collection_function):
	"""
	Get the order parameter, based on the CNA, for a cluster. This order parameter is the normalised similarity compared to the similarity of clusters in the population. 

	:param cluster: This is the cluster you want to obtain rho_i for.
	:type  cluster: Organisms.GA.Cluster
	:param max_similarity: The maximum similarity obtained from the population + offspring_pools
	:type  max_similarity: float
	:param min_minimarity: The minimum similarity obtained from the population + offspring_pools
	:type  min_minimarity: float
	:param cna_database: This is an in-memory database (not a ASE disk database) that records the similarities between clusters in the population.
	:type  cna_database: Organisms.GA.SCM_Scripts.CNA_Database
	:param cna_fitness_function: This is the function that converts the rho similarity into a fitness value. 
	:type  cna_fitness_function: __func__

	:returns: CNA_fitness_value: This is the SRM fitness contribution to be used to obtain the fitness valuee 
	:rtypes: float

	"""
	CNA_most_similar_average = get_CNA_most_similar_average(cluster,cna_database,collection_function)
	# Change this to an fitness parameter, where: # Changed 3/12/2018, GRW
	# 	* High similarity -> large fitness parameter -> less favourable for mating and natural selection
	# 	* Low similarity  -> small fitness parameter -> more favourable for mating and natural selection
	CNA_fitness_contribution = (CNA_most_similar_average - min_minimarity)/(max_similarity - min_minimarity) # Note from GRW. This term should be CNA_most_similar_average/100.0
	return CNA_fitness_contribution


def get_CNA_fitness_contribution_normalised(cluster,max_similarity,min_minimarity,cna_database,collection_function,cna_fitness_function):
	"""
	Get the fitness, based on the structural diversity as determined by the SCM, for a cluster. This order parameter is the normalised similarity compared to the similarity of clusters in the population. 

	:param cluster: This is the cluster you want to obtain rho_i for.
	:type  cluster: Organisms.GA.Cluster
	:param max_similarity: The maximum similarity obtained from the population + offspring_pools
	:type  max_similarity: float
	:param min_minimarity: The minimum similarity obtained from the population + offspring_pools
	:type  min_minimarity: float
	:param cna_database: This is an in-memory database (not a ASE disk database) that records the similarities between clusters in the population.
	:type  cna_database: Organisms.GA.SCM_Scripts.CNA_Database
	:param cna_fitness_function: This is the function that converts the rho similarity into a fitness value. 
	:type  cna_fitness_function: __func__

	"""
	rho_i = get_CNA_fitness_parameter_normalised(cluster,max_similarity,min_minimarity,cna_database,collection_function)
	fitness = cna_fitness_function.get_fitness(rho_i)
	return fitness, rho_i

# ------------------------------------------------------------------------------------------------------------------------------------------------

def get_lowest_and_highest_similarities_from_collections(population,collections,cna_database,collection_function):
	"""
	This method will return the value of the highest and lowest similarities from a list of collections that are inputted into this method.
	The collections variable is a list of collections, for example [population,offspring]
	This is a private method.

	:param collections: A list of all the collections that you want to compare..
	:type  collections: list of collection objects

	:returns: the lowest similarity of clusters out of all the inputed collections, the maximum similarity out of all the inputed collections.
	:rtypes: float, float

	"""
	cluster_names = population.get_cluster_names()
	if not collections == None:
		#if not len(collections[0]) == 16:
		#	import pdb; pdb.set_trace()
		for collection in collections:
			cluster_names += collection.get_cluster_names()
	
	lowest_similarity = 101.0
	highest_similarity =  -1.0
	for cluster_name in cluster_names:
		all_cna_averages = cna_database.get_all_averages_for_a_cluster(cluster_name)
		try:
			CNA_most_similar_average = float(collection_function(all_cna_averages))
		except:
			print('Weird CNA fitness contribution issue in CNA_Fitness_Contribution')
			import pdb; pdb.set_trace()
			exit('Exitting')
		if CNA_most_similar_average < lowest_similarity:
			lowest_similarity = CNA_most_similar_average
		if CNA_most_similar_average > highest_similarity:
			highest_similarity = CNA_most_similar_average
	if lowest_similarity == 101.0:
		exit('get_lowest_and_highest_similarities_from_collections, low error')
	if highest_similarity == -1.0:
		exit('get_lowest_and_highest_similarities_from_collections, high error')
	return lowest_similarity, highest_similarity



'''
Did_Find_LES_Main.py, Geoffrey Weal, 08/03/2019

This program Will determine which of your genetic algorithm trials have found the cluster with the energy given to a certain decimal place.
'''
import os, sys
import numpy as np
import scipy.stats

def Rounding_Method(number,rounding_criteria):
    """
    This definition will round the input number to the decimal place specified by rounding_criteria.
    The input number in the genetic algorithm is usually the energy of the cluster.

    :param number: The value that you would like to round (usually energy)
    :type  number: float
    :param rounding_criteria: The number of decimal places the value should be rounded to.
    :type  rounding_criteria: int
    
    """
    if rounding_criteria > 12:
        print('Error, the maximum decimal place rounding, to avoid numerical errors, is 12')
        print('rounding_criteria: '+str(rounding_criteria))
        exit('Check this out')
    number = round(number,rounding_criteria)
    return number

def mean_confidence_interval(data, confidence=0.95):
    a = 1.0 * np.array(data)
    n = len(a)
    m, se = np.mean(a), scipy.stats.sem(a)
    ci = se * scipy.stats.t.ppf((1 + confidence) / 2., n-1)
    return m, ci

class has_all_trials_found_LES:

    def __init__(self,energy_LES, max_generation_to_survey, set_rounding):
        self.energy_LES = Rounding_Method(float(energy_LES),set_rounding)
        self.max_generation_to_survey = max_generation_to_survey
        self.set_rounding = set_rounding
        self.run()
        self.print_those_that_all_trials_did_not_all_find_LES()

    def print_those_that_all_trials_did_not_all_find_LES(self):
        results_all_found_LES = []
        results_not_all_found_LES = []
        for dirpath, all_number_of_true_trials, all_total_no_of_trials in self.results:
            if all_number_of_true_trials == all_total_no_of_trials:
                results_all_found_LES.append((dirpath, all_number_of_true_trials, all_total_no_of_trials))
            else:
                results_not_all_found_LES.append((dirpath, all_number_of_true_trials, all_total_no_of_trials))
        print('# ------------------------------------------------------------------ #')
        if not results_all_found_LES == []:
            print('Jobs that all trials found the LES')
            for dirpath, all_number_of_true_trials, all_total_no_of_trials in results_all_found_LES: 
                print(str(dirpath)+' ('+str(all_number_of_true_trials)+' out of '+str(all_total_no_of_trials)+')')
        else:
            print('No Jobs were completed that all trials found the LES')
        print('# ------------------------------------------------------------------ #')
        if not results_not_all_found_LES == []:
            print('Jobs that NOT all trials found the LES')
            for dirpath, all_number_of_true_trials, all_total_no_of_trials in results_not_all_found_LES: 
                print(str(dirpath)+' ('+str(all_number_of_true_trials)+' out of '+str(all_total_no_of_trials)+')')
        else:
            print('For all Jobs, all trials found the LES')
        print('# ------------------------------------------------------------------ #')

    def run(self):
        path = os.getcwd()
        Overall_Trials_to_check = {}
        self.results = []
        for dirpath, dirnames, filenames in os.walk(path):
            dirnames.sort()
            if any([(dirname.startswith('Trial') and dirname.replace('Trial','').isdigit()) for dirname in dirnames]):
                print('Getting LES data for '+str(dirpath))
                LES_Details = self.get_found_LES_Details(dirpath, dirnames, self.energy_LES, self.max_generation_to_survey, self.set_rounding)
                toString = ''
                toString += '******************************************************************************\n'
                toString += '******************************************************************************\n'
                toString += 'This is the information of the LES of these genetic algorithm trials.\n'
                toString += 'Trial\tfound LES\tdir\tGen.\n'
                for LES_of_a_Trial in LES_Details:
                    LES_Trial, found_LES, LES_clu_dir, LES_gen = LES_of_a_Trial
                    toString += str(LES_Trial)+'\t'+str(found_LES)+'\t'+str(LES_clu_dir)+'\t'+str(LES_gen)+'\n'
                toString += '******************************************************************************\n'
                toString += '******************************************************************************\n'
                toString += 'Overall Details'+'\n'
                number_of_true_trials = [x for x in LES_Details if x[1] == True]
                number_of_true_trials = len(number_of_true_trials)
                total_no_of_trials = len(LES_Details)
                toString += 'Number of trials that located the LES: '+str(number_of_true_trials)+' out of '+str(total_no_of_trials)+'\n'
                print(toString)
                toString += '******************************************************************************\n'
                toString += '******************************************************************************\n'
                #with open(dirpath+'/DidFind_LESOfTrials.txt','w') as LESOfTrials:
                #    LESOfTrials.write(toString)
                name = dirpath.replace(path+'/','')
                name = name.replace('/','_')
                #with open(path+'/'+ name+'_DidFind_LESOfTrials.txt','w') as LESOfTrials:
                #    LESOfTrials.write(toString)
                dirnames[:] = []
                filenames[:] = []
                #self.make_file_of_data(dirpath,None,LES_Details,energy_LES)
                #self.make_file_of_data(path,name,LES_Details,energy_LES)
                self.results.append((dirpath, number_of_true_trials, total_no_of_trials))

        print('########################################################################')
        print('########################################################################')
        
        print('########################################################################')
        print('########################################################################')

    def get_found_LES_Details(self, dirpath, dirnames, energy_LES, max_generation_to_survey, rounding):

        def Lowest_Energy_from_Trial(filepath, energy_LES, max_generation_to_survey, rounding):
            #LES_clu_dir = -1; LES_gen = -1;
            LES_clu_dir = None; LES_gen = None;
            found_LES = False
            if os.path.exists(filepath):
                got_to_trial = False
                pop_size = 0
                no_of_off_per_gen = 0
                last_gen_counter = 0
                with open(filepath,'r') as DETAILS_FILE:
                    for line in DETAILS_FILE:
                        if line.startswith('Genetic Algorithm Starts Here.'):
                            continue
                        if line.startswith('Restarting due to epoch.'):
                            continue
                        if line.startswith('Finished prematurely as LES energy found.'):
                            got_to_trial = True
                            to_string = 'Error: You have found the message "Finished prematurely as LES energy found." in your energyprofile.txt, but you didnt locate the cluster energy you entered. Check that the energy you put in is the energy of the LES, and run this program again. This error occurred for dirpath: '+str(filepath)
                            raise Exception(to_string)
                            break
                        clu_dir, gen, energy = line.rstrip().split()
                        clu_dir = int(clu_dir); gen = int(gen)
                        if gen == 0:
                            pop_size += 1
                        if gen == 1:
                            no_of_off_per_gen += 1
                        if gen == max_generation_to_survey:
                            got_to_trial = True
                            last_gen_counter += 1
                        if gen > max_generation_to_survey:
                            break
                        energy = Rounding_Method(float(energy),rounding)
                        if energy == energy_LES:
                            LES_clu_dir = clu_dir; LES_gen = gen;
                            found_LES = True
                            got_to_trial = True
                            break
                        #if energy < energy_LES:
                        #    LES_clu_dir = clu_dir; LES_gen = gen; energy_LES = energy
                if got_to_trial == False:
                    print(filepath)
                    print('Error, 2. Gen: '+ str(gen))
                    import pdb; pdb.set_trace()
                    exit()
                if gen == max_generation_to_survey:
                    if last_gen_counter < no_of_off_per_gen:
                        exit('Error, 3')
            return found_LES, LES_clu_dir, LES_gen

        LES_Data_from_Trials = []
        dirnames = [dirname for dirname in dirnames if dirname.startswith('Trial')]
        dirnames.sort(key=lambda dirname: int(dirname.replace('Trial','')))
        number_of_dirnames = len(dirnames)
        for index in range(number_of_dirnames):
            dirname = dirnames[index]
            trial_no = int(dirname.replace('Trial',''))
            found_LES, LES_clu_dir, LES_gen = Lowest_Energy_from_Trial(dirpath+'/'+dirname+'/Population/EnergyProfile.txt',energy_LES,max_generation_to_survey,rounding)
            LES_Data_from_Trials.append([dirname, found_LES, LES_clu_dir, LES_gen])
            ###########
            sys.stdout.write("\r                                                                  ")
            sys.stdout.flush()
            sys.stdout.write("\rScanning Completion: "+str(float(index+1)/float(number_of_dirnames)*100.0)+" % (Checked "+str(dirname)+").")
            sys.stdout.flush()
            ###########
        LES_Data_from_Trials.sort(key=lambda x:x[1],reverse=False)
        return LES_Data_from_Trials

    def make_file_of_data(self, dirpath, name, LES_Details, energy_LES):

        if not name in ['', None]:
            name += '_'

        success_at_gens = [0]*(self.max_generation_to_survey+1)
        LES_Details.sort(key=lambda x:x[3],reverse=False)
        putative_LES_energy = LES_Details[0][3]
        for dirname, LES_clu_dir, LES_gen, LES_energy in LES_Details:
            if LES_energy == putative_LES_energy:
                success_at_gens[LES_gen] += 1

        with open(dirpath+'/'+str(name)+'first_instance_up_to_'+str(self.max_generation_to_survey)+'.txt','w') as first_instanceTXT:
            generation = 0
            for success_at_gen in success_at_gens:
                first_instanceTXT.write(str(generation)+': '+str(success_at_gen)+'\n')
                generation += 1

        success_at_gens_collaborated = [0]
        for success_at_gen in success_at_gens:
            at_gen_success = success_at_gens_collaborated[-1] + success_at_gen
            success_at_gens_collaborated.append(at_gen_success)
        del success_at_gens_collaborated[0]

        with open(dirpath+'/'+str(name)+'_did_find_LES_'+str(energy_LES)+'_up_to_'+str(self.max_generation_to_survey)+'.txt','w') as successTXT:
            generation = 0
            for success_at_gen_collaborated in success_at_gens_collaborated:
                successTXT.write(str(generation)+': '+str(success_at_gen_collaborated)+'\n')
                generation += 1




#!/usr/bin/python

EnergyProfileTXT_path = 'Population/EnergyProfile.txt'
up_to_name = 1
energies = []
restart_points = []
with open(EnergyProfileTXT_path,'r') as EnergyProfileTXT:
	for line in EnergyProfileTXT:
		if line.startswith('Genetic Algorithm Starts Here.'):
			continue
		if line.startswith('Restarting due to epoch.'):
			restart_points.append(clu_dir)
			continue
		if line.startswith('Finished prematurely as LES energy found.'):
			break
		clu_dir, gen, energy = line.rstrip().split()
		clu_dir = int(clu_dir); gen = int(gen); energy = float(energy)
		if not up_to_name == clu_dir:
			exit('Error')
		up_to_name += 1
		energies.append(energy)



import copy
from random import randrange, uniform, randint, random
from collections import Counter

from Organisms.GA.Cluster import Cluster
from Organisms.GA.ExternalDefinitions import InclusionRadiusOfCluster, is_position_already_occupied_by_an_atom_in_Cluster

def moveMutate(cluster_to_mutate, dist_to_move):
	'''
	This method randomly displaces all the atoms within the structure cluster_to_mutate from its original position.
	This def depends on the value of the maximum bond length of the cluster, self.r_ij.
	The def will cause all atoms in the cluster to move randomly by up to a 1/2 of self.r_ij from their original position.

	:param cluster_to_mutate: The cluster to move the atoms in the cluster.
	:type  cluster_to_mutate: GA.Cluster
	:param dist_to_move: The fistnace to move clusters by
	:type  dist_to_move: float

	:returns The newly created mutated cluster
	:rtypes  GA.Cluster
	'''
	mutant = cluster_to_mutate.deepcopy()
	# Randomly move each atom of the cluster
	for atom in mutant:
		distance_left_to_move = dist_to_move
		####################
		movement_x = uniform(-1.,1.)*distance_left_to_move
		atom.x += movement_x
		distance_left_to_move = (distance_left_to_move**2.0 - movement_x**2.0)**0.5
		####################
		movement_y = uniform(-1.,1.)*distance_left_to_move
		atom.y += movement_y
		distance_left_to_move = (distance_left_to_move**2.0 - movement_y**2.0)**0.5
		####################
		movement_z = distance_left_to_move
		atom.z += uniform(-1.,1.)*movement_z
		####################
	if not len(mutant) == len(cluster_to_mutate):
		print('Error in def moveMutate, in Types_Of_Mutations.py')
		print('The offspring contains '+str(len(mutant))+' atoms, but should contain '+str(len(cluster_to_mutate)))
		print('Check this')
		import pdb; pdb.set_trace()
		print('This program will finish without completing')
		exit()
	return mutant

def homotopMutate(cluster_to_mutate):
	'''
	This definition is designed to swap the positions of two elementally different atoms in a cluster
	
	:param cluster_to_mutate: The cluster to move the atoms in the cluster.
	:type  cluster_to_mutate: GA.Cluster

	:returns The newly created mutated cluster
	:rtypes  GA.Cluster

	'''
	# randomly select a atom in the cluster
	atom1Index = randrange(0,len(cluster_to_mutate))
	# randomly select a second atom in the cluster which is not the same as the first and is of a different element compared to the first atom. 
	while True:
		atom2Index = randrange(0,len(cluster_to_mutate))
		# If the two atoms picked are the same or are the same type of atom, try again (can make this more efficient)
		if not (cluster_to_mutate[atom1Index].symbol == cluster_to_mutate[atom2Index].symbol or atom1Index == atom2Index):
			break
	# Swap element types of pair in terms of positions. Note (GRW) I did originally just swap
	# the symbol variable in mutant but thought there are alot of other variables I am not thinking about
	# transferring as well which may cause issues. Swapping the positions seemed less likely to bring up
	# problems in the future.
	print('Atom '+str(atom1Index)+' ('+str(cluster_to_mutate[atom1Index].symbol)+') will be swapped with atom '+str(atom2Index)+' ('+str(cluster_to_mutate[atom2Index].symbol)+').')
	mutant = copy.deepcopy(cluster_to_mutate)
	temp = copy.deepcopy(mutant[atom1Index].position)
	mutant[atom1Index].position = copy.deepcopy(mutant[atom2Index].position)
	mutant[atom2Index].position = copy.deepcopy(temp)
	return mutant

def randomMutate(boxtoplaceinlength,vacuumAdd,cluster_makeup=None,cluster_to_mutate=None,percentage_of_cluster_to_randomise=None):
	"""
	This definition provides the random method for the mutation proceedure. In this method, a cluster is
	designed by randomly placing the designed atoms of elements into a predefined unit cell.
	
	:param boxtoplaceinlength: This is the length of the box you would like to place atoms in to make a randomly constructed cluster.
	:type  boxtoplaceinlength: float
	:param vacuumAdd: The length of vacuum added around the cluster. Written in A.
	:type  vacuumAdd: float
	:param cluster_makeup: check this
	:type  cluster_makeup: {'Element': int(number of that 'element' in this cluster),...}
	:param cluster_to_mutate: If the user desired, they can tell this definition what cluster they want to mutate. Default: None.
	:type  cluster_to_mutate: ASE.Atoms
	:param percentage_of_cluster_to_randomise: This is the percentage of the number of atoms in the cluster to randomise
	:type  percentage_of_cluster_to_randomise: float
	
	:returns mutant: The description of the mutant cluster.
	:rtypes: ase.Atoms or GA.Cluster

	"""
	# Prepare the method for performing a fully ramdomly generated cluster
	if not cluster_makeup == None and (cluster_to_mutate == None or percentage_of_cluster_to_randomise == None):
		#Turn cluster_makeup into a chemical formula string
		cluster_chemical_formula = '' 
		for element, no_of_element in cluster_makeup.items():
			cluster_chemical_formula += str(element+str(no_of_element))
		# set up the cluster for randomizing
		mutant = Cluster(cluster_chemical_formula); nAtoms = len(mutant); atoms_to_randomise = range(nAtoms) # randomise all atoms in the cluster.
	# preparing this method for only randomly changing the position of only a percentage (of atoms) of a cluster.
	elif cluster_makeup == None and not (cluster_to_mutate == None or percentage_of_cluster_to_randomise == None):
		mutant = copy.deepcopy(cluster_to_mutate); nAtoms = len(mutant)
		# The following will pick random atoms in the cluster to randomise
		no_of_atoms_to_randomise = int(ceil(float(nAtoms)*(float(percentage_of_cluster_to_randomise)/100.0)))
		all_atoms_in_cluster = range(nAtoms); atoms_to_randomise = []
		for NOTUSED in range(no_of_atoms_to_randomise):
			index_to_randomise = randint(0,len(all_atoms_in_cluster))-1
			atoms_to_randomise.append(all_atoms_in_cluster.pop(index))
	else: # Error.
		print('Error in MutationProcedure: def randomMutate.')
		print('cluster_makeup = ' + str(cluster_makeup))
		print('cluster_to_mutate = ' + str(cluster_to_mutate))
		print('percentage_of_cluster_to_randomise = ' + str(percentage_of_cluster_to_randomise))
		import pdb; pdb.set_trace()
		exit()
	# At this point, the method needs:
	#	* mutant: class ASE.Atoms. A framework of the cluster.
	#	* atoms_to_randomise: the index of atoms that need to be randomised, list of ints
	# setup the Atoms class to define the atoms in the cluster of mutant and define a unit cell
	# for atoms to be randomly placed in.
	# Get initial size of cell\
	mutant.set_cell([boxtoplaceinlength,boxtoplaceinlength,boxtoplaceinlength])
	mutant.center()
	# For each element, place a set number of atoms of that element into the cluster in a random position
	# within the defined cube with sides = r_ij*scale
	for index in atoms_to_randomise:
		while True:
			x_position = boxtoplaceinlength*uniform(0,1)
			y_position = boxtoplaceinlength*uniform(0,1)
			z_position = boxtoplaceinlength*uniform(0,1)
			position = [x_position, y_position, z_position]
			# check the cluster to see that this new random position will not cause this atom 
			# to be in the same place as any atom in the cluster currently
			if not is_position_already_occupied_by_an_atom_in_Cluster(position,mutant,atom_indices_to_exclude_from_comparison=[index]):
				# Found that one of the atoms in this cluster has the same position as this randomly generated one.
				# Will need to make a new randomly generated position list
				break
		mutant[index].position = position
	# Add a unit cell to the cluster. Mainly needed for VASP.
	lengthOfCell = 2.0*InclusionRadiusOfCluster(mutant) + vacuumAdd
	cell = [lengthOfCell,lengthOfCell,lengthOfCell]
	mutant.set_cell(cell)
	mutant.center() # centre the cluster within this unit cell
	# return the mutant cluster.
	#print('cluster_makeup: '+str(cluster_makeup)+'; cluster_to_mutate: '+str(cluster_to_mutate))
	if cluster_makeup == None: 
		if not len(mutant) == len(cluster_to_mutate):
			print('Error in def randomMutate, in Types_Of_Mutations.py')
			print('The offspring contains '+str(len(mutant))+' atoms, but should contain '+str(len(cluster_to_mutate)))
			print('Check this')
			import pdb; pdb.set_trace()
			print('This program will finish without completing')
			exit()
	elif cluster_to_mutate == None:
		number_of_atoms = sum(Counter(cluster_makeup).values())
		#print('Atoms in mutant: '+str(len(mutant))+'; number_of_atoms = '+str(number_of_atoms))
		#import pdb; pdb.set_trace()
		if not len(mutant) == number_of_atoms:
			print('Error in def randomMutate, in Types_Of_Mutations.py')
			print('The offspring contains '+str(len(mutant))+' atoms, but should contain '+str(len(cluster_to_mutate)))
			print('Check this')
			import pdb; pdb.set_trace()
			print('This program will finish without completing')
			exit()
	return mutant
		


'''
Eenrgy_Predation_Operator.py, Geoffrey Weal, 28/10/2018

This is one of the Predation Operators that can be used by the genetic algorithm program. 

'''

from Organisms.GA.Predation_Operators.Predation_Operator import Predation_Operator

class Energy_Predation_Operator(Predation_Operator):
	"""
	This is one of the Predation Operators that can be used by the genetic algorithm program. 

	This class will allow the genetic algorithm to procedure without any Predation Operator. Instead, this class provide a way to use
	the energetic fitness to obtain a fitness for each cluster. This Operator will not remove any clusters from the population or 
	the offspring pool.

	:param Predation_Information: This contains all the information needed by the Predation Operator you want to use to run.
	:type  Predation_Information: dict.
	:param population: This is the population that this Operator will be controlling to make sure that no two clusters in the population have the same energy.
	:type  population: Organisms.GA.Population
	:param print_details: Print details of the predation operator, like verbose
	:type  print_details: bool

	"""
	def __init__(self,predation_information,population,print_details):
		super().__init__(predation_information,population,print_details)
		self.Energy_Predation_Operators_Options(predation_information)
		self.get_energy_predation_methods()

	def Energy_Predation_Operators_Options(self,predation_information):
		"""
		This method is designed to add and check the options that you have placed in to 
		the predation_information dictionary for the Energy Predation Operator

		:param predation_information: This contains all the information needed by the Predation Operator you want to use to run.
		:type  predation_information: dict.

		"""
		if not predation_information['Predation Operator'] == 'Energy':
			print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
			print('The Predation Operator "Energy_Predation_Operator" has been initialised.')
			print("However, predation_information['Predation_Operator'] is not 'Energy'")
			print("predation_information['Predation_Operator'] = "+str(predation_information['Predation_Operator']))
			print('Check this out.')
			import pdb; pdb.set_trace()
			exit('This program will exit without completing.')
		try:
			self.mode = predation_information['mode']
		except:
			print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
			print("You need to include a 'mode' in your predation_information dictionary.")
			print("This should be either 'simple' or 'comprehensive'.")
			print('Do this and repeat this algorithm.')
			print('The GA Program will end.')
			exit('This program will exit without completing.')
		############################################################
		if self.mode == 'simple':
			try:
				self.round_energy = int(predation_information['round_energy'])
			except ValueError:
				print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
				print("predation_information input 'round_energy' but be intergerable.")
				print("predation_information['round_energy'] = "+str(predation_information['round_energy']))
				print('Do this and repeat this algorithm.')
				print('The GA Program will end.')
				exit('This program will exit without completing.')
			except:
				print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
				print("You need to set a 'round_energy' in to your predation_information variable in your Run.py (or MakeTrial.py) file, as you have set mode = 'simple'.")
				print("E.g. predation_information['round_energy'] = 2")
				print('Do this and repeat this algorithm.')
				print('The GA Program will end.')
				exit('This program will exit without completing.')
			try:
				predation_information['minimum_energy_diff']
				print("Ignoring your 'minimum_energy_diff' input variable for the Energy Predation Operator")
				del predation_information['minimum_energy_diff']
			except:
				pass
		elif self.mode == 'comprehensive':
			try:
				self.type_of_comprehensive_scheme = predation_information['type_of_comprehensive_scheme']
			except:
				print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
				print("You need to set a 'type_of_comprehensive_scheme' in to your predation_information variable in your Run.py (or MakeTrial.py) file, as you have set mode = 'comprehensive'.")
				print("E.g. predation_information['type_of_comprehensive_scheme'] = 'energy' or predation_information['type_of_comprehensive_scheme'] = 'fitness'")
				print('Do this and repeat this algorithm.')
				print('The GA Program will end.')
				exit('This program will exit without completing.')

			try:
				self.minimum_energy_diff = predation_information['minimum_energy_diff']
			except ValueError:
				print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
				print("predation_information input 'minimum_energy_diff' but be floatable.")
				print("predation_information['minimum_energy_diff'] = "+str(predation_information['minimum_energy_diff']))
				print('Do this and repeat this algorithm.')
				print('The GA Program will end.')
				exit('This program will exit without completing.')
			except:
				print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
				print("You need to set a 'minimum_energy_diff' in to your predation_information variable in your Run.py (or MakeTrial.py) file, as you have set mode = 'comprehensive'.")
				print("E.g. predation_information['minimum_energy_diff'] = 0.005 #eV")
				print('Do this and repeat this algorithm.')
				print('The GA Program will end.')
				exit('This program will exit without completing.')
			try:
				predation_information['round_energy']
				print("Ignoring your 'round_energy' input variable for the Energy Predation Operator")
				del predation_information['round_energy']
			except:
				pass
		else:
			print('Error in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
			print("The variable 'mode' in your predation_information dictionary should be either 'simple' or 'comprehensive'.")
			print('Do this and repeat this algorithm.')
			print('The GA Program will end.')
			exit('This program will exit without completing.')
		############################################################

	def get_energy_predation_methods(self):
		"""
		This method is designed to import the methods needed to run this Predation Operator, using with the 'simple' or 'comprehensive' mode.
		"""
		if self.mode == 'simple':
			from Organisms.GA.Predation_Operators.Energy_Predation_Operator_Scripts.Simple_Energy_Predation_Scheme import check_initial_population, assess_for_violations
		elif self.mode == 'comprehensive':
			if self.type_of_comprehensive_scheme == 'energy':
				from Organisms.GA.Predation_Operators.Energy_Predation_Operator_Scripts.Comprehensive_Energy_Predation_Operator_energy  import check_initial_population, assess_for_violations_force_replacement, assess_for_violations_no_force_replacement
			elif self.type_of_comprehensive_scheme == 'fitness':
				from Organisms.GA.Predation_Operators.Energy_Predation_Operator_Scripts.Comprehensive_Energy_Predation_Operator_fitness import check_initial_population, assess_for_violations_force_replacement, assess_for_violations_no_force_replacement
			else:
				print('Error in def get_energy_predation_methods, in class Energy_Predation_Operator, in Energy_Predation_Operator.py')
				print('self.type_of_comprehensive_scheme must be either "energy" or "fitness".')
				print('Currently, self.type_of_comprehensive_scheme = '+str(self.type_of_comprehensive_scheme))
				print('Check this and change as appropriate.')
				exit('This program will exit without completing.')
		self.check_initial_population_imported = check_initial_population
		self.assess_for_violations_force_replacement_imported = assess_for_violations_force_replacement
		self.assess_for_violations_no_force_replacement_imported = assess_for_violations_no_force_replacement
		############################################################
			
	def check_initial_population(self,return_report=False):
		"""
		This definition is responsible for making sure that the initialised population obeys the Predation Operator of interest.

		:param return_report: Will return a dict with all the information about what clusters are similar to what other clusters in the population. 
		:type  return_report: bool.

		returns:
			* **clusters_to_remove** (*list of ints*): a list of the clusters to remove from the population as they violate the Predation Operator. Format is [(index in population, name fo cluster),...]
			* **CNA_report** (*dict.*): a dictionary with information on the clusters being removed and the other clusters in the population which have caused the violation to the SCM Predation Operator. This information is only used to display information so they know why there are violations to the Predation Operator when they occur. For is {removed cluster: [list of clusters that this cluster is similar to in the population.]}

		"""
		return self.check_initial_population_imported(self, return_report)

	def assess_for_violations(self,offspring_pool,force_replace_pop_clusters_with_offspring):
		"""
		This definition is designed to determine which offspring (and the clusters in the population) violate the predation Operator. 
		It will not remove or change any clusters in the offspring or population, but instead will record which offspring violate the 
		predation Operator. 

		It will also recommend which clusters in the population should be removed and be replaced by which offspring prior to the 
		natural selection process. Here, the cluster in the population and the offspring will be in violation of each other, however it 
		may be advantageous to keep the offspring rather than the cluster in the population as the offspring is fitter than the cluster
		in the offspring

		:param offspring_pool: This is the collection of offspring to assess for violations to the Predation Operator.
		:type  offspring_pool: Organisms.GA.Offspring_Pool.Offspring_Pool
		:param force_replace_pop_clusters_with_offspring: This will tell the genetic algorithm whether to swap clusters in the population with offspring if the predation operator indicates they are the same but the predation operator has a better fitness value than the cluster in the population. 
		:type  force_replace_pop_clusters_with_offspring: bool.

		returns:
			* **offspring_to_remove** (*tuple of ints*): A list of the names of the offspring to be removed
			* **force_replacement** (*tuple of (int, int)*): A list of the clusters in the population that should be replaced, and the offspring they should be replaced by.

		"""
		if force_replace_pop_clusters_with_offspring:
			return self.assess_for_violations_force_replacement_imported(self, offspring_pool)
		else:
			return self.assess_for_violations_no_force_replacement_imported(self, offspring_pool)

	########################################################################################################
	########################################################################################################
	########################################################################################################

	def add_to_database(self, collection):
		'''
		Add clusters similarities to the CNA database to be stored for future generations. 

		:param collection: update the fitnesses of clusters in the collection.
		:type  collection: Organisms.GA.Collection.Collection
		'''
		pass

	def remove_from_database(self, cluster_names_to_remove):
		'''
		Clusters to remove from the CNA database

		:param cluster_names_to_remove: A list of the names of all the clusters to remove from the CNA database
		:type  cluster_names_to_remove: list of ints
		'''
		pass

	def reset(self):
		'''
		Reset the CNA database with no inputs
		'''
		pass


# ---------------------------------------------------------------------------------------------------
def get_resume_from_generation(self):
	"""
	Will determine what the clusters in the population are, and the current generation, if the Organisms program is being restarted or if the number of generations is being increased.

	Will get this information for files in the Population folder.

	:returns: current_generation: The current generation that the Organisms program is being restarted from, or otherwise None if this is a new Organisms program run.
	:rtype: int or None
	:returns: clusters: The clusters in the population. In the formation of [(name of cluster, energy of cluster)]
	:rtype: [(int, float), ...]

	"""
	current_generation, clusters, cluster_energies, got_information_from_backup_state_file = self.population.get_current_generation_from_state_file()
	return current_generation, clusters, cluster_energies, got_information_from_backup_state_file
	#current_generation, clusters = self.population.get_current_generation_from_collection_history()
	#return current_generation, clusters

def Initial_ProgramChecking(self):
	"""
	This definition will check data from files and from that obtained self.get_resume_from_generation() and check to whether
	the program can continue or not.
	"""
	def EndProgram():
		print('The program will end')
		exit()
	if self.resume_from_generation == None:
		return
	if self.resume_from_generation >= self.generations:
		print('The Genetic Algorithm program has already completed the maximum number of generations desired.')
		EndProgram()

# ---------------------------------------------------------------------------------------------------
from Organisms.GA.Initialise_Population import Initialise_Population
def Initate_New_GAProgram(self):
	"""
	Set up the Organisms program to perform a new genetic algorithm run.

	returns: the name of the last cluster that was created.
	rtype: int
	"""
	previous_cluster_name = 0
	generation = 0
	self.epoch.setting_up_for_new_GA()
	self.memory_operator.setup_from_database(0,self.resume_from_generation)
	last_cluster_number = Initialise_Population(self.population,self.cluster_makeup,self.surface,self.Minimisation_Function,self.memory_operator,self.predation_operator,self.fitness_operator,self.epoch,self.cell_length,self.vacuum_to_add_length,self.r_ij,self.rounding_criteria,self.no_of_cpus,previous_cluster_name=previous_cluster_name,generation=generation)
	self.ga_recording_system.record_initial_populations(self.population) # self.ga_recording_system.record_initial_populations(self.population)
	self.ga_recording_system.add_collection(self.population,[])
	self.energyprofile.add_collection(self.population,generation)
	return last_cluster_number

# ---------------------------------------------------------------------------------------------------
import os, sys, time
from Organisms.GA.GA_Program_external_methods import convert_seconds_to_normal_time
def Resume_GAProgram(self, resume_from_generation, clusters_in_resumed_population, clusters_in_resumed_population_energies):
	"""
	This definition initalises all python variables and rewrites and modifies all files to the point when the
	generation self.resume_from_generation finished and the next generation began.

	:param resume_from_generation: the generation the genetic algorithm is resuming from. 
	:type  resume_from_generation: int
	:param clusters_in_resumed_population: list of the name of the clusters in the population to resume from. 
	:type  clusters_in_resumed_population: list of int
	:param clusters_in_resumed_population_energies: list of the name of the clusters in the population to resume from 
	:type  clusters_in_resumed_population_energies: list of float

	returns: the name of the last cluster that was created.
	rtype:   int

	"""
	self.epoch.setting_up_epoch_to_resume_GA()
	if not self.population.have_database:
		print('Error: You are wanting to resume this genetic algorithm, however the population is not recording a database to store the last generation of clusters')
		print('The genetic algorithm is unable to resume without the database of clusters.')
		print('Check this out and restart this genetic algorithm from beginning.')
		exit('This program will finish without completing')
	if not (self.population.does_contain_database(backup=False) or self.population.does_contain_database(backup=True)):
		print('Error: You are wanting to resume this genetic algorithm.')
		print('However, there is no database in the Population folder.')
		print('The genetic algorithm is unable to resume without the database of clusters.')
		print('Check this out and restart this genetic algorithm from beginning.')
		exit('This program will finish without completing')
	if not self.epoch.information_from_the_same_generation(resume_from_generation):
		print('Error in resuming the genetic algorithm')
		print('The current generation is: '+str(resume_from_generation))
		print('However, the information from the epoch is from generation: '+str(self.epoch.initial_generation))
		print('Check this')
		exit('This program will end without doing anything')
	print('----------------------------------------------')
	print('Loading Population stored on disk into program')
	print('----------------------------------------------')
	used_backup = self.population.import_clusters_from_database_to_memory(resume_from_generation, clusters_in_resumed_population, clusters_in_resumed_population_energies, self.rounding_criteria)
	self.population.check_historyfile(resume_from_generation)
	##################################################################################################
	# create folders EnergyProfile and PoolHistory with included files, PoolProfile.txt and EnergyProfile.txt files and get them ready to use.
	self.energyprofile.check(resume_from_generation, self.no_offspring_per_generation)
	# assign fitnesses to the clusters in the population
	self.fitness_operator.assign_resumed_population_fitnesses(resume_from_generation=resume_from_generation)
	##################################################################################################
	# get previous cluster name
	generation, last_cluster_generated_name = self.energyprofile.get_current_generation_and_last_cluster_generated_from_EnergyProfile()
	if not resume_from_generation == generation:
		print('Error at def Resume_GAProgram in GA_Program, in GA_Program.py')
		print('Check')
		import pdb; pdb.set_trace()
		exit()
	##################################################################################################
	# Check that the population obeys the predation operator
	print('----------------------------------------------')
	print('Checking for any predation violation between clusters in the population.')
	print('----------------------------------------------')
	clusters_removed_from_pop, removed_clusters_report = self.predation_operator.check_initial_population(return_report=True)
	if len(clusters_removed_from_pop):
		print('Error in def Place_Already_Created_Clusters_In_Population, in GA_Initiate.py.')
		print('Some of the clusters that you have placed in the population folder violate the Predation Operator: '+str(Fitness_Factor_Operator.Predation_Switch))
		print('Here is a list of the clusters that are the same as other clusters in this population:')
		for cluster_kept_dir, clusters_removed_dirs in sorted(removed_clusters_report.iteritems(),key=lambda x:x[0]):
			clusters_removed_dirs.insert(0,cluster_kept_dir)
			print('Clusters: '+str(clusters_removed_dirs)+' have the same energy.')
		print('Check this out.')
		print('This program will end without doing anything')
		exit()
	##################################################################################################
	# Set up the memory_operator
	self.memory_operator.setup_from_database(last_cluster_generated_name,resume_from_generation)
	for cluster in self.population:
		if self.memory_operator.is_similar_cluster_in_memory_operator(cluster):
			print('Error in def Place_Already_Created_Clusters_In_Population, in GA_Initiate.py.')
			exit('This program will end without doing anything')
	##################################################################################################
	# Check the GA_Recording_Database database
	print('----------------------------------------------')
	print('Checking the GA_Recording_Database database.')
	print('Note that if you have a big database with over 100,000 files, this may take a bit of time.')
	start_checking_ga_recording_system_database = time.time()
	self.ga_recording_system.resume_ga_recording_system_from_current_generation(resume_from_generation)
	end_checking_ga_recording_system_database = time.time()
	print('Checking the GA_Recording_System database took '+convert_seconds_to_normal_time(end_checking_ga_recording_system_database-start_checking_ga_recording_system_database)+'.')
	print('----------------------------------------------')
	##################################################################################################
	if used_backup:
		self.population.move_backup_database_to_normal_backup()
	else:
		self.population.remove_backup_database_if_exists()
	print('----------------------------------------------')
	print('Finished resetting the population.')
	print('Will continue from generation '+str(resume_from_generation+1))
	print('----------------------------------------------')
	return last_cluster_generated_name

# ---------------------------------------------------------------------------------------------------
def add_metadata(self):
	"""
	This is included as this seems to not be added to the collections database when it is first created. Fixing an issue with ASE.
	"""
	self.ga_recording_system.add_metadata()
	self.population.add_metadata()

# ---------------------------------------------------------------------------------------------------
from Organisms.GA.GA_Program_Details import GA_Program_Details
def GA_Initiate(self):
	"""
	This definition provides the details to how to run the genetic algorithm. The program has the ability to restart if required.
	"""
	########################################################
	# Will find out if the program has been run previously and if so, provide details about the generation, the clusters and their energies in the population to resume from.
	#self.resume_from_generation, clusters_in_resumed_population = get_resume_from_generation(self)
	self.resume_from_generation, clusters_in_resumed_population, clusters_in_resumed_population_energies, got_information_from_backup_state_file = get_resume_from_generation(self)
	Initial_ProgramChecking(self)
	#import pdb; pdb.set_trace()
	########################################################
	# Set the Organisms program up so that it ready to begin a new genetic algorithm run, or continue the original genetic algorithm run.
	if self.resume_from_generation == None:
		self.previous_cluster_name = Initate_New_GAProgram(self)
		self.starting_generation = 0
		self.ga_program_details = GA_Program_Details(self, True) # Set up the GA details program to write information about how performance of the genetic algorithm
	elif self.resume_from_generation >= 0: # Continue from the generation that you left off from.
		#import pdb; pdb.set_trace()
		#self.previous_cluster_name, self.resume_from_generation = Resume_GAProgram(self,self.resume_from_generation,clusters_in_resumed_population)
		self.previous_cluster_name = Resume_GAProgram(self,self.resume_from_generation,clusters_in_resumed_population,clusters_in_resumed_population_energies)
		if got_information_from_backup_state_file == True:
			self.population.move_backup_to_current_files()
		self.starting_generation = self.resume_from_generation
		self.ga_program_details = GA_Program_Details(self, False) # Set up the GA details program to write information about how performance of the genetic algorithm
	else:
		exit('Error with resuming genetic algorithm in GA_Initiate.py, def GA_Initiate(self)')
	# The population has been initised either from anew or from some generation, perform the offspring generation and natural selection processes.
	self.starting_generation += 1
	add_metadata(self)
	print('Time elapsed: '+str(self.timer.print_elapsed_time()))
	print('Date and Time: '+str(self.timer.get_time_now()))



#!/usr/bin/python
'''
Geoffrey Weal, mass_scancel.py, 10/02/2021

This program is designed to cancel all the jobs between id_low and id_high in slurm.
'''
import os, sys
from subprocess import call, Popen

id_low = int(sys.argv[1])
id_high = int(sys.argv[2])

for scancel_id in range(id_low,id_high+1):
	args = ["scancel",str(scancel_id)]
	print('Cancelling Job id: '+str(scancel_id))
	p=Popen(args)


import os
import time
from datetime import datetime

class Timer:
	"""
	This class is designed to time the genetic algorithm, as well as to determine if the algorithm has run for longer than is desired by the total_length_of_running_time variable.

	:param total_length_of_running_time: This is the length of time to run the algorithm for before safely finishing the algorithm. The time is given in hours
	:type  total_length_of_running_time: float
	"""
	def __init__(self,total_length_of_running_time):
		self.total_length_of_running_time = total_length_of_running_time # Hours
		if not self.total_length_of_running_time == None:
			self.total_length_of_running_time *= (60.0*60.0) # turns the time in to seconds. 
		self.start_time = time.time()

	def get_total_length_of_running_time(self):
		'''
		Return the maximum amount of time to run the genetic algorithm for, in hours.

		:returns: total_length_of_running_time, the maximum amount of time to run the genetic algorithm for, in hours.
		:rtype:   float or str.
		'''
		if self.total_length_of_running_time == None: 
			return 'Indefinitely'
		else:
			return str(self.total_length_of_running_time/(60.0*60.0))+' hours'

	def has_elapsed_time(self):
		"""
		This determines if the cluster has exceeded the desired running time.
		"""
		if self.total_length_of_running_time == None:
			return False
		elapsed_time = time.time() - self.start_time
		return (elapsed_time > self.total_length_of_running_time)

	def print_elapsed_time(self):
		"""
		Returns the current running time of the algorithm
		"""
		elapsed_time = time.time() - self.start_time
		return time.strftime("%H:%M:%S", time.gmtime(elapsed_time))

	def get_time_now(self):
		"""
		Returns the current date and time.
		"""
		return datetime.now()


#!/usr/bin/python
'''
HasCompletedUpToGeneration.py, Geoffrey Weal, 08/03/2019

This program will determine if your genetic algorithm trials has completed up to a certain generation. 

This program will also determine how many genetic algorithm trials have found th eputative global minimum from across all the trials, as well as the average number of minimisations required to obtain the putative global minimum up to the generation specified

If you want to get the most complete average number of generations required to obtain the putative global minimum, make sure you run all your genetic algorithm trials so that they have all found the true global minimum, then set the generation to search up to to infinity (you do not need to enter a generation, the default setting is infinity).

'''
import sys
from Organisms.Postprocessing_Programs.GetLESOfAllTrials_Main_v3 import GetLESOfAllTrials_Main

################################################################################################################################
################################################################################################################################
################################################################################################################################
################################################################################################################################
################################################################################################################################

def get_input_int(input_message,default_input):
    if not isinstance(default_input,int):
        print('Error in get_input_int of Coagulate_EnergyProfiles.py')
        print('The default_input must be an int')
        print('default_input: '+str(default_input))
        print('Check this out')
        import pdb; pdb.set_trace()
        exit()
    while True:
        get_input = str(raw_input(str(input_message)+' ['+str(default_input)+']?: '))
        get_input.lower()
        if get_input == '':
            get_input = str(default_input)
        if not get_input.isdigit():
            print('Error. Your input must be an integer.')
        else:
            return int(get_input)

def get_int(given_input):
    given_input = str(given_input)
    given_input.lower()
    if not given_input.isdigit():
        print('Error. Your input must be an integer.')
    else:
        return int(given_input)

number_of_sys_argvs = len(sys.argv)

if number_of_sys_argvs == 1:
    max_generation_to_survey = get_input_int('What generation would you like to survey up to?',sys.maxsize)
    set_rounding = get_input_int('What rounding of the energy would you like to set?',12)
elif number_of_sys_argvs == 2:
    max_generation_to_survey = get_int(sys.argv[1])
    print('Generation to survey up to: '+str(max_generation_to_survey))
    set_rounding = get_input_int('What rounding of the energy would you like to set?',12)
elif number_of_sys_argvs == 3:
    max_generation_to_survey = get_int(sys.argv[1])
    print('Generation to survey up to: '+str(max_generation_to_survey))
    set_rounding = get_int(sys.argv[2])
    print('Rounding of the energy to set: '+str(set_rounding))

GetLESOfAllTrials_Main(max_generation_to_survey,set_rounding)


import os

from Organisms.GA.Surface import Surface

from Organisms.GA.Population import Population
from Organisms.GA.Offspring_Pool import Offspring_Pool

from Organisms.GA.Crossover import Crossover
from Organisms.GA.Mutation import Mutation

from Organisms.GA.Memory_Operator import Memory_Operator
from Organisms.GA.Get_Predation_and_Fitness_Operators import get_predation_and_fitness_operators
from Organisms.GA.Epoch import Epoch

from Organisms.GA.GA_Recording_System import GA_Recording_System
from Organisms.GA.EnergyProfile import EnergyProfile
from Organisms.GA.Timer import Timer

def GA_Setup(self,cluster_makeup,pop_size,generations,no_offspring_per_generation,creating_offspring_mode,crossover_type,mutation_types,
	chance_of_mutation,r_ij,vacuum_to_add_length,Minimisation_Function,surface_details,epoch_settings,cell_length,memory_operator_information,
	predation_information,fitness_information,ga_recording_information,force_replace_pop_clusters_with_offspring,user_initialised_population_folder,
	rounding_criteria,print_details,no_of_cpus,finish_algorithm_if_found_cluster_energy,total_length_of_running_time):
	"""
	This method will set up the genetic algorithm.

	:param cluster_makeup: This contains the information on the makeup of the cluster you would like to optimise for. Format is a dictionary in the form of: {element: number of that element}
	:type  cluster_makeup: {str: int, ...}
	:param pop_size: The size of the population
	:type  pop_size: int
	:param generations: The number of generations that are run
	:type  generations: int
	:param no_offspring_per_generation: The number of offspring that are created per generation
	:type  no_offspring_per_generation: int
	:param creating_offspring_mode: This indicates how the offspring are created, either via the mating method 'followed' by the mutation method, or by only perform the mating method 'or' mutation method, or (i.e. either mating and/or mutation). See manual for how to set this.
	:type  creating_offspring_mode: str.
	:param crossover_type: This is the type of crossover that you would like to use. See the manual for more information.
	:type  crossover_type: str.
	:param mutation_types: This is a list that contains all the information about the mutation methods you would like to use. 
	:type  mutation_types: list of (str., float)
	:param chance_of_mutation: This indicates the change of a mutation occuring. See the manual on specifically how this works.
	:type  chance_of_mutation: float
	:param r_ij: This is the maximum bond distance that we would expect in this cluster. See the manual for more information. 
	:type  r_ij: float
	:param vacuum_to_add_length: This is the amount of vacuum to place around the cluster.
	:type  vacuum_to_add_length: float
	:param Minimisation_Function: This is a function that determines how to locally minimise clusters. See manual for more information.
	:type  Minimisation_Function: __func__
	:param surface_details: This functionality has not been designed yet. Default: None
	:type  surface_details: None
	:param epoch_settings: This is designed to hold the information about the epoch method. 
	:type  epoch_settings: dict.
	:param cell_length: This is the length of the cubic unit cell to construct clusters in. See manual for more information. Default: 'default'
	:type  cell_length: float
	:param predation_information: This holds all the information about the predation operator. Default: {'Predation Operator':"Off"}
	:type  predation_information: dict.
	:param fitness_information: This holds all the information about the fitness operator. Default: {'Fitness Operator':"Off"}
	:type  fitness_information: dict.
	:param ga_recording_information: Default: {}
	:type  ga_recording_information: dict.
	:param force_replace_pop_clusters_with_offspring: This will tell the genetic algorithm whether to swap clusters in the populatino with offspring if the predation operator indicates they are the same but the predation operator has a better fitness value than the cluster in the population. 
	:type  force_replace_pop_clusters_with_offspring: bool.
	:param user_initialised_population_folder: This is the directory to a folder containing any custom made clusters you would like to include in the initial population. Set this to None if you do not have any initial clusters to add into the population. Default: None
	:type  user_initialised_population_folder: str. or None
	:param rounding_criteria: The number of decimal places to round the energies of clusters made during the genetic algorithm to. Default: 2
	:type  rounding_criteria: int
	:param print_details: Verbose for this algorithm.
	:type  print_details: bool.
	:param no_of_cpus: The number of cpus that the algorithm can use via multiprocessing. Default: 1
	:type  no_of_cpus: int
	:param finish_algorithm_if_found_cluster_energy: If desired, the algorithm can finish if the LES is located. This is useful to use for methods testing. The algorithm will determine that the LES is found when the genetic algorithm locates the energy of the LES. Read the manual on how to use this. Default: None
	:type  finish_algorithm_if_found_cluster_energy: dict. or None
	:param total_length_of_running_time: The total amount of time to run the genetic algorithm for. If the algorithm is still running after this time, the algorithm will safety finish. Time given in hours. None means no limit on time, Default: None.
	:type  total_length_of_running_time: int or None

	"""
	self.timer = Timer(total_length_of_running_time)
	self.Run_path = os.getcwd()
	# set_number_of_cpus
	self.no_of_cpus = no_of_cpus
	##########################################################################################################
	################################# Variables for the Genetic Algorithm ####################################
	##########################################################################################################
	# This details the elemental and number of atom composition of cluster that the user would like to investigate
	self.cluster_makeup = cluster_makeup
	# Surface Details
	self.surface_details = surface_details
	self.surface = Surface(surface_details)
	self.place_cluster_where = None
	# These are the main variables of the genetic algorithm that with changes could affect the results of the Genetic Algorithm.
	self.population_name = 'Population'
	self.pop_size = pop_size
	self.user_initialised_population_folder = user_initialised_population_folder
	# ------------------------------------------------------------------------------------------- #
	self.population = Population(self.population_name,self.pop_size,user_initialised_population_folder=self.user_initialised_population_folder)
	self.energyprofile = EnergyProfile(self.population)
	if self.energyprofile.is_LES_note_in_EnergyProfile():
		print("Found the note, 'Finished prematurely as LES energy found.', in the populations energyprofile.")
		print('This means you have located the LES that you desired.')
		print('Program will end')
		exit()
	# ------------------------------------------------------------------------------------------- #
	#self.energyprofile_backup = EnergyProfile(self.population)
	self.offspring_pool_name = 'Offspring'
	self.generations = generations
	self.no_offspring_per_generation = no_offspring_per_generation
	self.offspring_pool = Offspring_Pool(self.offspring_pool_name,self.no_offspring_per_generation)
	# These are variables used by the algorithm to make and place clusters in. Variables of the cell to generate offspring
	# ------------------------------------------------------------------------------------------- #
	self.r_ij = r_ij
	self.cell_length = cell_length
	if self.cell_length in ['default','Default','DEFAULT']:
		self.cell_length = self.r_ij * (sum([float(noAtoms) for noAtoms in self.cluster_makeup.values()]) ** (1.0/3.0))
	elif isinstance(self.cell_length,float):
		pass
	else:
		print('Error in your GA Run.py script.')
		print('Your "cell_length" variable is not a float, nor is "default"')
		print('cell_length = '+str(self.cell_length))
		print('Check this. THis program will end without doing anything.')
		exit()
	self.vacuum_to_add_length = vacuum_to_add_length
	# These setting indicate how offspring should be made using the Mating and Mutation Proceedures
	self.creating_offspring_mode = creating_offspring_mode 
	self.crossover_type = crossover_type
	self.mutation_types = mutation_types
	self.chance_of_mutation = chance_of_mutation
	self.crossover_procedure = Crossover(self.crossover_type,self.r_ij,self.vacuum_to_add_length,sum(self.cluster_makeup.values()))
	self.mutation_procedure  = Mutation(self.mutation_types,self.r_ij,self.vacuum_to_add_length)
	# ------------------------------------------------------------------------------------------- #
	# This object will remember any clusters to avoid in the population
	self.memory_operator_information = memory_operator_information
	self.memory_operator = Memory_Operator(self.memory_operator_information)
	# ------------------------------------------------------------- #
	# This contains all the informatino about the predation operator and fitness operator that will be used.
	self.predation_information = predation_information 
	self.fitness_information = fitness_information
	self.predation_operator, self.fitness_operator = get_predation_and_fitness_operators(self.predation_information, self.fitness_information, self.population, self.generations, self.no_of_cpus, False)
	self.force_replace_pop_clusters_with_offspring = force_replace_pop_clusters_with_offspring
	# ------------------------------------------------------------------------------------------- #
	# Create and set the conditions for performing an epoch
	self.epoch = Epoch(epoch_settings,self.Run_path,self.fitness_operator,self.population)
	# ------------------------------------------------------------- #
	# The RunMinimisation.py algorithm is one set by the user. It contain the def Minimisation_Function
	# That is used for local optimisations. This can be written in whatever way the user wants to perform
	# the local optimisations. This is meant to be as free as possible.
	# Variable which provides the directory for the script or program for running the local
	# optimisation method.
	import inspect
	if not inspect.isfunction(Minimisation_Function):
		exit('Error: Minimisation_Function must be a function/def (definition)')
	self.Minimisation_Function = Minimisation_Function
	# These are last techinical points that the algorithm is designed in mind
	if rounding_criteria > 12:
		print('Error, the maximum decimal place rounding, to avoid numerical errors, is 12.')
		print('rounding_criteria: '+str(rounding_criteria))
		exit('Check this out. This program will finish without completing.')
	self.rounding_criteria = rounding_criteria
	self.similarity_rounding_criteria = 10
	# Variables required for the Recording_Cluster.py class/For recording the history as required of the genetic algorithm.
	# For recording the details of the clusters. Recording_Clusters is designed to 
	# keep a history of clusters that have been created during the genetic algorithm.
	self.ga_recording_information = ga_recording_information
	self.ga_recording_system = GA_Recording_System(self.ga_recording_information)
	# For Info.txt
	self.noOfForceCalls = []
	self.timeForForceCalls = []
	##########################################################################################################
	# Set for debugging options
	self.print_details = print_details
	#self.debug = True
	##########################################################################################################
	if finish_algorithm_if_found_cluster_energy == None:
		self.finish_algorithm_if_found_cluster_energy = None
	else:
		if finish_algorithm_if_found_cluster_energy['cluster energy'] == None:
			self.finish_algorithm_if_found_cluster_energy = None
		else:
			self.finish_algorithm_if_found_cluster_energy = finish_algorithm_if_found_cluster_energy['cluster energy']
			self.finish_algorithm_if_found_cluster_rounding = finish_algorithm_if_found_cluster_energy['round']


#!/usr/bin/python
'''
Postprocessing_Database.py, Geoffrey Weal, 26/9/2020

This program is designed to break down the database from the GA_Recording_System into manageable chunks if the original database was too large to process.
'''
import os, sys
import numpy as np
from ase.db.row import row2dct
from ase.db import connect

def get_distance(atom1,atom2):
	diff_x = atom1.x - atom2.x
	diff_y = atom1.y - atom2.y
	diff_z = atom1.z - atom2.z
	distance = (diff_x**2.0 + diff_y**2.0 + diff_z**2.0)**0.5
	return distance

def get_cluster_interatomic_distances(cluster):
	cluster_interatomic_distances = {}
	for index1 in range(len(cluster)):
		for index2 in range(index1+1,len(cluster)):
			cluster_interatomic_distances[(index1,index2)] = round(get_distance(cluster[index1],cluster[index2]),4)
	return cluster_interatomic_distances

def compared_cluster_interatomic_distances(cluster_interatomic_distances_1, cluster2):
	for index1 in range(len(cluster2)):
		for index2 in range(index1+1,len(cluster2)):
			cluster2_interatomic_distance = round(get_distance(cluster2[index1],cluster[index2]),4)
			cluster1_interatomic_distance = cluster_interatomic_distances_1[(index1,index2)]
			if not cluster2_interatomic_distance == cluster1_interatomic_distance:
				exit('Error')

def get_rotation_matrix(i_v, unit=None):
    # From http://www.j3d.org/matrix_faq/matrfaq_latest.html#Q38
    if unit is None:
        unit = (0.0,0.0,1.0)
    # Normalize vector length
    i_v /= np.linalg.norm(i_v)
    # Get axis
    uvw = np.cross(i_v, unit)
    # compute trig values - no need to go through arccos and back
    rcos = np.dot(i_v, unit)
    rsin = np.linalg.norm(uvw)
    #normalize and unpack axis
    if not np.isclose(rsin, 0):
        uvw /= rsin
    u, v, w = uvw
    # Compute rotation matrix - re-expressed to show structure
    return (
        rcos * np.eye(3) +
        rsin * np.array([
            [ 0, -w,  v],
            [ w,  0, -u],
            [-v,  u,  0]
        ]) +
        (1.0 - rcos) * uvw[:,None] * uvw[None,:]
    )

def rotate_cluster_to_major_inertia_vector(cluster):
	"""
	This method will rotate its axis such that the cluster 
	"""
	evals, evecs = cluster.get_moments_of_inertia(vectors=True)
	major_evec = evecs[0]
	matrix = get_rotation_matrix(major_evec)
	cluster.center(about=(0., 0., 0.))
	cluster_interatomic_distances_initial = get_cluster_interatomic_distances(cluster)
	cluster.set_positions(np.dot(matrix,cluster.get_positions().T).T)
	compared_cluster_interatomic_distances(cluster_interatomic_distances_initial, cluster)
	cluster.center()
	return cluster

# ----------------------------------------------------------------------------------------------------------------------------

sort_cluster_keys = ['id', 'name', 'cluster_energy']

def get_input_int(input_message,default_input):
	if not isinstance(default_input,int):
		print('Error in get_input_int of Coagulate_EnergyProfiles.py')
		print('The default_input must be an int')
		print('default_input: '+str(default_input))
		print('Check this out')
		import pdb; pdb.set_trace()
		exit()
	while True:
		get_input = str(raw_input(str(input_message)+' ['+str(default_input)+']?: '))
		get_input.lower()
		if get_input == '':
			get_input = str(default_input)
		if not get_input.isdigit():
			print('Error. Your input must be an integer.')
		else:
			return int(get_input)

def get_int(given_input):
	given_input = str(given_input)
	given_input.lower()
	if not given_input.isdigit():
		print('Error. Your input must be an integer.')
	else:
		return int(given_input)

def get_input_str(input_message,default_input):
	if not isinstance(default_input,str):
		print('Error')
		print('The default_input must be an str')
		print('default_input: '+str(default_input))
		print('Check this out')
		import pdb; pdb.set_trace()
		exit()
	while True:
		return str(raw_input(str(input_message)+' ['+str(default_input)+']?: '))

number_of_sys_argvs = len(sys.argv)

if number_of_sys_argvs == 1:
	number_of_clusters_per_database = get_input_int('How many clusters would you like in each database?',sys.maxsize)
	sort_clusters_by = get_input_str('What parameter would you like the clusters in the database(s) sorted by? [options: '+str(sort_cluster_keys)+']','cluster_energy')
elif number_of_sys_argvs == 2:
	number_of_clusters_per_database = get_int(sys.argv[1])
	print('No. of clusters in each database: '+str(number_of_clusters_per_database))
	sort_clusters_by = get_input_str('What parameter would you like the clusters in the database(s) sorted by? [options: '+str(sort_cluster_keys)+']','cluster_energy')
elif number_of_sys_argvs == 3:
	number_of_clusters_per_database = get_int(sys.argv[1])
	print('No. of clusters in each database: '+str(number_of_clusters_per_database))
	sort_clusters_by = str(sys.argv[2])
	print('Sorting clusters by: '+str(sort_clusters_by))

if not sort_clusters_by in sort_cluster_keys:
	print('Error in Postprocessing_Database.py')
	print('Your input for sort_clusters_by must be one of either: '+str(sort_cluster_keys))
	print('Your input for sort_clusters_by: '+str(sort_clusters_by))
	print('Check this out.')
	exit('This program will finish without completing.')

# ----------------------------------------------------------------------------------------------------------------------------

database_name = 'GA_Recording_Database'
if os.path.exists(database_name+'.db'):
	db = connect(database_name+'.db')
else:
	exit('Error')

for file in os.listdir(os.getcwd()):
	if file.startswith(database_name+'_postprocessed_') and file.endswith('.db'):
		os.remove(file)

clusters_info = []

for row in db.select():
	if not (sort_clusters_by in row._keys):
		exit('Error')
	print(row.id)
	clusters_info.append((row.id, row.get(sort_clusters_by)))

clusters_info.sort(key=lambda x:x[1])

counter = 1
database_number = 1
db_smaller = connect(database_name+'_postprocessed_'+str(database_number)+'.db')
for cluster_id, cluster_datum in clusters_info:
	if counter == number_of_clusters_per_database+1:
		counter = 1
		database_number += 1
		db_smaller = connect(database_name+'_postprocessed_'+str(database_number)+'.db')
	row = db.get(id=cluster_id)
	cluster = row.toatoms()
	cluster = rotate_cluster_to_major_inertia_vector(cluster)
	cluster_data = {}
	for key in row._keys:
		cluster_data[key] = row[key]
	db_smaller.write(cluster, key_value_pairs=cluster_data)
	print(counter)
	counter += 1




import os 
from shutil import copyfile

from Organisms.GA.Initialise_Population import Initialise_Population
from Organisms.GA.Get_Predation_and_Fitness_Operators import get_fitness_operator

class Epoch:
	"""
	This class is responsible for the epoch method used during the genetic algorithm run

	:param epoch_settings: This is a dictionary of all the epoch settings.
	:type  epoch_settings: dict.
	:param path: This is the path to store the epoch information in after each generation has completed. This information allows the genetic algorithm run to continue with information of the epoch from the last completed generation. 
	:type  path: str.

	"""
	def __init__(self,epoch_settings,path,fitness_operator,population):
		self.path = path
		self.epoch_data_name = 'epoch_data'
		self.backup_epoch_data_name = self.epoch_data_name+'.backup'
		self.epoch_settings = epoch_settings
		self.set_settings(self.epoch_settings,fitness_operator,population)

	def __repr__(self):
		return str(self.__dict__)

	def set_settings(self,epoch_settings,fitness_operator,population):
		"""
		This sets up the epoch as desired. 

		:param epoch_settings: This is a dictionary of all the epoch settings.
		:type  epoch_settings: dict.

		"""
		if epoch_settings == None:
			epoch_settings = {'epoch mode': None}
		self.epoch_mode = epoch_settings['epoch mode']
		# If there is no epoch
		if self.epoch_mode in ['Off', 'off', 'None', 'none', None, False]:
			self.epoch_function = self.no_epoch
			self.perform_epoch_function = self.perform_epoch_no_epoch
			return
		elif self.epoch_mode in ['mean energy']:
			self.energy_difference = epoch_settings['mean energy difference']
			self.epoch_function = self.should_epoch_using_mean_energy_epoch
			self.perform_epoch_function = self.perform_epoch_mean_energy
		elif self.epoch_mode in ['same population']:
			self.max_number_of_times_non_changing_population = epoch_settings['max repeat']
			self.epoch_function = self.should_epoch_using_same_population_epoch
			self.perform_epoch_function = self.perform_epoch_same_population
		else:
			print('Issue with the Epoch method, def set_settings in Epoch.py')
			print('You must set the epoch method to one of the following:')
			print('  --> Off: No epoch method')
			print('  --> mean energy: epoch method based on the mean energy of the population')
			print('  --> same population: epoch method based on repeated clusters in the population')
			print('See the manual for more details.')
			exit('This program will exit without finishing')
		# if epoching first by changing fitness function to energy fitness function
		first_epoch_to_change_fitness_function_name = 'first epoch changes fitness operator to energy fitness operator'
		if first_epoch_to_change_fitness_function_name in epoch_settings:
			self.first_epoch_to_change_fitness_function = epoch_settings[first_epoch_to_change_fitness_function_name]
			if self.first_epoch_to_change_fitness_function:
				if fitness_operator.fitness_switch == 'Energy':
					print('Error with the Epoch method setting: '+str(first_epoch_to_change_fitness_function_name))
					print('You have chosen that the first time a population epochs that it changes the fitness function to the Energy fitness function')
					print('But you are already using the Energy fitness function.')
					print('This will be ineffective and could cause issues. Therefore we are stopping the program here so you can sort this out')
					print(first_epoch_to_change_fitness_function_name+': '+str(self.first_epoch_to_change_fitness_function))
					print('fitness_operator.fitness_switch: '+str(fitness_operator.fitness_switch))
					exit('This program will finish without completing.')
				elif fitness_operator.fitness_switch == 'SCM + Energy' and float(fitness_operator.SCM_fitness_contribution) == 0.0: 
					print('Error with the Epoch method setting: '+str(first_epoch_to_change_fitness_function_name))
					print('You have chosen that the first time a population epochs that it changes the fitness function to the Energy fitness function')
					print('But you are using the SCM + Energy fitness function with SCM_fitness_contribution = 0.0')
					print('This is equivalent to the Energy fitness function.')
					print('This will be ineffective and could cause issues. Therefore we are stopping the program here so you can sort this out')
					print(first_epoch_to_change_fitness_function_name+': '+str(self.first_epoch_to_change_fitness_function))
					print('fitness_operator.fitness_switch: '+str(fitness_operator.fitness_switch))
					exit('This program will finish without completing.')
				self.perform_reset_population = True
				#if fitness_operator.fitness_switch == 'Energy':
				#	self.fitness_information_temp = {'Fitness Operator': 'Energy', 'fitness_function': fitness_operator.fitness_information['fitness_function']}
				if fitness_operator.fitness_switch == 'SCM + Energy':
					self.fitness_information_temp = {'Fitness Operator': 'Energy', 'fitness_function': epoch_settings['energy_fitness_function']} #'fitness_function': fitness_operator.fitness_information['energy_fitness_function']}
				else:
					exit('fitness_operator switch error')
				predation_information = {'Predation Operator': 'Off'}
				self.fitness_operator_temp = get_fitness_operator(self.fitness_information_temp, None, population, None, 1, False)
		else:
			self.first_epoch_to_change_fitness_function = False

	# --------------------------------------------------------------------------------------------------------
	# --------------------------------------------------------------------------------------------------------
	# Get data for new genetic algorithm trial

	def setting_up_for_new_GA(self):
		if self.epoch_mode in ['mean energy']:
			self.initialise_mean_energy_epoch_details()
		# import data if same population epoch method was used
		elif self.epoch_mode in ['same population']:
			self.initialise_same_population_epoch_details()
		else:
			print('Error in Epoch')
			print('You have not chosen an available epoch method')
			print('epoch parameter "epoch mode" should be either: "None", "mean energy", or "same population"')
			print("Your Epoch's 'epoch mode': "+str(self.epoch_mode))
			print('Check this')
			exit('This program will exit without completing')
		del self.epoch_settings

	def initialise_mean_energy_epoch_details(self):
		"""
		This contains the information required to set up the mean_energy epoch method. 

		:param epoch_settings: This is a dictionary of all the epoch settings.
		:type  epoch_settings: dict.

		"""
		self.initial_generation = 0
		self.previous_mean_energy = float('inf')

	def initialise_same_population_epoch_details(self):
		"""
		This contains the information required to set up the same_population epoch method. 

		:param epoch_settings: This is a dictionary of all the epoch settings.
		:type  epoch_settings: dict.

		"""
		self.initial_generation = 0
		self.clusters_in_previous_population = None
		self.number_of_times_non_changing_population = 0

	# --------------------------------------------------------------------------------------------------------
	# --------------------------------------------------------------------------------------------------------
	# Get data for resumed genetic algorithm trial

	def setting_up_epoch_to_resume_GA(self):
		"""
		This method will load the data from the epoch data on disk in order to resume.
		"""
		# check whether to use information from the epoch data on disk or from the backup epoch data.
		# If an epoch method was used, make sure that data is obtained from the backup file before obtaining data from the original file. 
		if self.does_backup_exist()               and self.is_epoch_data_of_disk_complete(self.backup_epoch_data_name,self.epoch_settings):
			get_epoch_data_from = self.backup_epoch_data_name
			self.got_data_from_backup = True
		elif self.does_epoch_data_exist_on_disk() and self.is_epoch_data_of_disk_complete(self.epoch_data_name,self.epoch_settings):
			get_epoch_data_from = self.epoch_data_name
			self.got_data_from_backup = False
		else:
			print('Error with the Epoch data on disk')
			print('The epoch data and the backup file of this are either not there, or is incomplete.')
			print('Make sure that at least one of the epoch data or its backup exist and is complete for the epoch method that you are using')
			print('See the manual for information on the information that is required for the epoch data to be complete.')
			exit('This method will finish without having begun.')
		'''
		# If an epoch method was used, make sure that data is obtained from the backup file before obtaining data from the original file. 
		if self.does_backup_exist():
			get_epoch_data_from = self.backup_epoch_data_name
			self.got_data_from_backup = True
		else:
			get_epoch_data_from = self.epoch_data_name
			self.got_data_from_backup = False
		'''
		# import data if mean energy epoch method was used
		if self.epoch_mode in ['mean energy']:
			self.get_resumed_mean_energy_epoch_details(self.epoch_settings,get_epoch_data_from)
		# import data if same population epoch method was used
		elif self.epoch_mode in ['same population']:
			self.get_resumed_same_population_epoch_details(self.epoch_settings,get_epoch_data_from)
		else:
			print('Error in Epoch')
			print('You have not chosen an available epoch method')
			print('epoch parameter "epoch mode" should be either: "None", "mean energy", or "same population"')
			print("Your Epoch's 'epoch mode': "+str(self.epoch_mode))
			print('Check this')
			exit('This program will exit without completing')
		# make backups
		if self.did_get_data_from_backup():
			self.replace_with_backup()
		del self.epoch_settings

	def get_resumed_mean_energy_epoch_details(self,epoch_settings,get_epoch_data_from):
		"""
		This contains the information required to set up the mean_energy epoch method. 

		:param epoch_settings: This is a dictionary of all the epoch settings.
		:type  epoch_settings: dict.
		:param get_epoch_data_from: The path of the epoch document to get information from if the genetic algorithm is being resumed.
		:type  get_epoch_data_from: str.

		"""
		#self.energy_difference = epoch_settings['mean energy difference']
		#if os.path.exists(get_epoch_data_from):
		with open(get_epoch_data_from,'r') as epoch_data:
			self.initial_generation = eval(epoch_data.readline())
			self.previous_mean_energy = eval(epoch_data.readline())
		#else:
		#	self.initial_generation = 0
		#	self.previous_mean_energy = float('inf')
		#self.epoch_function = self.should_epoch_using_mean_energy_epoch
		#self.perform_epoch_function = self.perform_epoch_mean_energy

	def get_resumed_same_population_epoch_details(self,epoch_settings,get_epoch_data_from):
		"""
		This contains the information required to set up the same_population epoch method. 

		:param epoch_settings: This is a dictionary of all the epoch settings.
		:type  epoch_settings: dict.
		:param get_epoch_data_from: The path of the epoch document to get information from if the genetic algorithm is being resumed.
		:type  get_epoch_data_from: str.

		"""
		#if os.path.exists(get_epoch_data_from):
		with open(get_epoch_data_from,'r') as epoch_data:
			self.initial_generation = eval(epoch_data.readline())
			self.clusters_in_previous_population = eval(epoch_data.readline())
			self.number_of_times_non_changing_population = eval(epoch_data.readline())
		#else:
		#	self.initial_generation = 0
		#	self.clusters_in_previous_population = None
		#	self.number_of_times_non_changing_population = 0
		#self.max_number_of_times_non_changing_population = epoch_settings['max repeat']
		#self.epoch_function = self.should_epoch_using_same_population_epoch
		#self.perform_epoch_function = self.perform_epoch_same_population

	def is_epoch_data_of_disk_complete(self,get_epoch_data_from,epoch_settings):
		# import data if mean energy epoch method was used
		try:
			if self.epoch_mode in ['mean energy']:
				self.get_resumed_mean_energy_epoch_details(epoch_settings,get_epoch_data_from)
			# import data if same population epoch method was used
			elif self.epoch_mode in ['same population']:
				self.get_resumed_same_population_epoch_details(epoch_settings,get_epoch_data_from)
			# If you get to this point, the right epoch method contains the information required, so return True
			return True
		except:
			# Some information is missing from the epoch_data file, so return False
			return False
	# --------------------------------------------------------------------------------------------------------
	# --------------------------------------------------------------------------------------------------------
	# Should en epoch occur methods

	def should_epoch(self,collection,generation_number):
		"""
		Determines if the genetic algorithm should epoch.

		:param collection: This is the collection to use to determine if the genetic algorithm should epoch. This Collection should be the Population.
		:type  collection: Organisms.GA.Collection
		:param generation_number: The number of the last successful generation
		:type  generation_number: int

		returns Should an epoch occur. True if yes, False if no.
		rtype   bool.

		"""
		return self.epoch_function(collection, generation_number)

	def no_epoch(self,collection,generation_number):
		"""
		The proceedure to determine if an epoch should occur.

		Since no epoch is included in this algorithm, return False to mean do not perform an epoch

		:param collection: This is the collection to use to determine if the genetic algorithm should epoch. This Collection should be the Population.
		:type  collection: Organisms.GA.Collection
		:param generation_number: The number of the last successful generation
		:type  generation_number: int

		returns Should an epoch occur. False for no, since this genetic algorithm has not been told to epoch.
		rtype   bool.
		"""
		return False

	def should_epoch_using_mean_energy_epoch(self,collection,generation_number):
		"""
		The proceedure to determine if an epoch should occur, using the mean_energy epoch method.

		:param collection: This is the collection to use to determine if the genetic algorithm should epoch. This Collection should be the Population.
		:type  collection: Organisms.GA.Collection
		:param generation_number: The number of the last successful generation
		:type  generation_number: int

		returns Should an epoch occur. True if yes, False if no.
		rtype   bool.
		"""
		mean_energy = collection.mean_energy()
		with open(self.epoch_data_name,'w') as epoch_data:
			epoch_data.write(str(generation_number)+'\n')
			epoch_data.write(str(mean_energy))
		if (self.previous_mean_energy - mean_energy) > self.energy_difference:
			self.previous_mean_energy = mean_energy
			return False
		else:
			self.previous_mean_energy = float('inf')
			return True

	def should_epoch_using_same_population_epoch(self,collection,generation_number):
		"""
		The proceedure to determine if an epoch should occur, using the same_population epoch method.

		Since no epoch is included in this algorithm, return False to mean do not perform an epoch

		:param collection: This is the collection to use to determine if the genetic algorithm should epoch. This Collection should be the Population.
		:type  collection: Organisms.GA.Collection
		:param generation_number: The number of the last successful generation
		:type  generation_number: int

		returns Should an epoch occur. True if yes, False if no.
		rtype   bool.
		"""
		clusters_in_population = collection.get_cluster_names()
		#print(clusters_in_population)
		if not clusters_in_population == self.clusters_in_previous_population:
			self.clusters_in_previous_population = list(clusters_in_population)
			self.number_of_times_non_changing_population = 0
		else:
			self.number_of_times_non_changing_population += 1
			if self.number_of_times_non_changing_population >= self.max_number_of_times_non_changing_population:
				return True
		with open(self.epoch_data_name,'w') as epoch_data:
			epoch_data.write(str(generation_number)+'\n')
			epoch_data.write(str(self.clusters_in_previous_population)+'\n')
			epoch_data.write(str(self.number_of_times_non_changing_population))
		return False

	# ----------------------------------------------------------------------
	# If to change the fitness function if to energy fitness function if first time the epoch was reached

	def change_fitness_function(self,fitness_operator):
		"""

		"""
		fitness_operator, self.fitness_operator_temp = self.fitness_operator_temp, fitness_operator
		self.perform_reset_population = not self.perform_reset_population
		return self.perform_reset_population, fitness_operator

	# ----------------------------------------------------------------------
	# Perform epoch by resetting the population

	def perform_epoch(self,generation_number,population,energyprofile,population_reset_settings,epoch_due_to_population_energy_convergence):
		"""
		Perform an epoch by resetting all the clusters in the population.

		:param generation_number: The number of the last successful generation
		:type  generation_number: int
		:param population: The number of the last successful generation
		:type  population: Organisms.GA.Population
		:param energyprofile: The energyprofile to write epoch information to.
		:type  energyprofile: Organisms.GA.EnergyProfile
		:param population_reset_settings: This is a dict that contains all the information requied to epoch the population with a set of randomly generated clusters. 
		:type  population_reset_settings: dict.
		:param epoch_due_to_population_energy_convergence: Did the Epoch occur because the population converged? True if yes, False if no. 
		:type  epoch_due_to_population_energy_convergence: bool.

		returns The name of the more recently generated cluster before the epoch method was performed.
		rtype   bool.
		"""
		cluster_makeup, surface, Minimisation_Function, memory_operator, predation_operator, fitness_operator, epoch_method, cell_length, vacuum_to_add_length, r_ij, rounding_criteria, no_of_cpus, previous_cluster_name = population_reset_settings	
		# ---------------------------------------------------------------------------------------- #
		# reset the predation and fitness operators
		fitness_operator.reset()
		predation_operator.reset()
		if epoch_due_to_population_energy_convergence:
			print('------------------------------------------------------------------------------------------')
			print('The clusters in the population all have the same energy, will need to reset the population')
			print('------------------------------------------------------------------------------------------')
			energyprofile.add_epoch_note_due_to_population_energy_convergence()
		else:
			energyprofile.add_epoch_note()
		# ---------------------------------------------------------------------------------------- #
		# reset the population
		for index in range(len(population)-1,-1,-1):
			population.remove(index)
		previous_cluster_name = Initialise_Population(population,cluster_makeup,surface,Minimisation_Function,memory_operator,predation_operator,fitness_operator,epoch_method,cell_length,vacuum_to_add_length,r_ij,rounding_criteria,no_of_cpus,previous_cluster_name=previous_cluster_name,generation=generation_number,get_already_created_clusters=False,is_epoch=True,epoch_due_to_population_energy_convergence=epoch_due_to_population_energy_convergence)
		# ---------------------------------------------------------------------------------------- #
		# reset the epoch_data file
		self.clusters_in_previous_population = population.get_cluster_names()
		self.number_of_times_non_changing_population = 0
		self.perform_epoch_function(generation_number)
		# ---------------------------------------------------------------------------------------- #
		energyprofile.add_collection(population,generation_number)
		return previous_cluster_name

	def perform_epoch_no_epoch(self,generation_number):
		"""
		This method is needed, but does not do anything.

		:param generation_number: The number of the last successful generation
		:type  generation_number: int
		"""
		return

	def perform_epoch_mean_energy(self, generation_number):
		"""
		This method writes the current epoch data to file for the mean_energy epoch method. 

		:param generation_number: The number of the last successful generation
		:type  generation_number: int
		"""
		mean_energy = collection.mean_energy()
		with open(self.epoch_data_name,'w') as epoch_data:
			epoch_data.write(str(generation_number)+'\n')
			epoch_data.write(str(mean_energy))

	def perform_epoch_same_population(self, generation_number):
		"""
		This method writes the current epoch data to file for the same_population epoch method. 

		:param generation_number: The number of the last successful generation
		:type  generation_number: int
		"""
		with open(self.epoch_data_name,'w') as epoch_data:
			epoch_data.write(str(generation_number)+'\n')
			epoch_data.write(str(self.clusters_in_previous_population)+'\n')
			epoch_data.write(str(self.number_of_times_non_changing_population))

	def does_epoch_data_exist_on_disk(self):
		"""
		Determine if a backup of the epoch information stored on disk exists.

		returns If a backup of the epoch information stored on disk exists. True for yes, False for no.
		rtype   bool.
		"""
		if self.epoch_mode == None:
			return
		return os.path.exists(self.path+'/'+self.epoch_data_name)

	# ----------------------------------------------------------------------
	# Epoch backup data

	def information_from_the_same_generation(self, generation_number):
		"""
		Determine if the epoch file contains the relavant information for the generation to resume from.

		:param generation_number: The number of the last successful generation
		:type  generation_number: int

		returns If the epoch information from the epoch file is for the current generation. True if yes, False if no. This value is True if no epoch method is used, as we dont need to deal with any epoch information if no epoch is used. 
		rtype   bool.
		"""
		if self.epoch_mode == None:
			return True
		return self.initial_generation == generation_number

	def did_get_data_from_backup(self):
		"""
		Determine if the data was obtained from backup epoch file or not. True for yes, False for no. If no epoch method was used, return False, since we dont need to deal with previous epoch information if no epoch method was used. 
	
		returns Did the epoch information come from the backup epoch file or not. True for yes, False for no. If no epoch method was used, return False, since we dont need to deal with previous epoch information if no epoch method was used.  
		rtype   bool.
		"""
		if self.epoch_mode == None:
			return False
		return self.got_data_from_backup

	def backup(self):
		"""
		Make a backup of the epoch data stored on disk
		"""
		if self.epoch_mode == None:
			return
		copyfile(self.path+'/'+self.epoch_data_name, self.path+'/'+self.backup_epoch_data_name)

	def remove_backup(self):
		"""
		Remove the backup of the epoch data stored on disk
		"""
		if self.epoch_mode == None:
			return
		os.remove(self.path+'/'+self.backup_epoch_data_name)

	def does_backup_exist(self):
		"""
		Determine if a backup of the epoch information stored on disk exists.

		returns If a backup of the epoch information stored on disk exists. True for yes, False for no.
		rtype   bool.
		"""
		if self.epoch_mode == None:
			return
		return os.path.exists(self.path+'/'+self.backup_epoch_data_name)

	def check_for_backup(self):
		"""
		Checks to see that a backup exists. If not, something has gone wrong and the algorithm needs to stop. 
		"""
		if self.epoch_mode == None:
			return
		if not os.path.exists(self.path+'/'+self.backup_epoch_data_name):
			print('Error in def replace_with_backup, in class Epoch, in Epoch.py')
			print('Could not find the epoch backup file. Called: '+str(self.backup_epoch_data_name))
			print('This should be found in '+str(self.path))
			print('Check this.')
			print('This program will finish without completing')
			exit()

	def replace_with_backup(self):
		"""
		Replace the current epoch data with the backup epoch data.
		"""
		if self.epoch_mode == None:
			return
		self.check_for_backup()
		os.remove(self.path+'/'+self.epoch_data_name)
		copyfile(self.path+'/'+self.backup_epoch_data_name,self.path+'/'+self.epoch_data_name)
		os.remove(self.path+'/'+self.backup_epoch_data_name)










#!/usr/bin/python
'''
Did_Find_LES.py, Geoffrey Weal, 08/03/2019

This program Will determine which of your genetic algorithm trials have found the cluster with the energy given to a certain decimal place.
'''
import os, sys
from Organisms.Postprocessing_Programs.Did_Find_LES_Main import has_all_trials_found_LES

path = os.getcwd()

# ---------------------------------------------------------------------- #
import re
def get_input_energy_float(input_message):
    while True:
        get_input = str(raw_input(str(input_message)+' [No default]?: '))
        if re.match(r'^-?\d+(?:\.\d+)?$', get_input) is None:
            print('Error. Your input must be an float.')
        try:
            return float(get_input)
        except:
            print('Error. Your input must be an integer.')

def get_input_int(input_message,default_input):
    if not isinstance(default_input,int):
        print('Error in get_input_int of Coagulate_EnergyProfiles.py')
        print('The default_input must be an int')
        print('default_input: '+str(default_input))
        print('Check this out')
        import pdb; pdb.set_trace()
        exit()
    while True:
        get_input = str(raw_input(str(input_message)+' ['+str(default_input)+']?: '))
        get_input.lower()
        if get_input == '':
            get_input = str(default_input)
        if not get_input.isdigit():
            print('Error. Your input must be an integer.')
        else:
            return int(get_input)

energy_LES = get_input_energy_float('What is the energy of the LES?')
set_rounding = get_input_int('What rounding of the energy would you like to set?',12)
max_generation_to_survey = get_input_int('What generation would you like to survey up to?',sys.maxsize)

# ---------------------------------------------------------------------- #

has_all_trials_found_LES(energy_LES, max_generation_to_survey, set_rounding)


#!/usr/bin/python
'''
Geoffrey Weal, view_cluster_from_database.py, 10/02/2021

This program is designed to show the user a cluster from the database with the desired id.
'''
import sys
from ase.io import read
from ase.visualize import view

id_name = sys.argv[1]

database_system_name = 'GA_Recording_Database.db'

system = read(database_system_name+'@id='+str(id_name))
if len(system) > 1:
	exit('Error')
system = system[0]
view(system)


import os
from ase.db import connect
from ase.io import read

from Organisms.GA.SCM_Scripts.TC_SCM_Methods import get_total_CNA_profile  as get_total_CNA_profile_T_SCM
from Organisms.GA.SCM_Scripts.AC_SCM_Methods import get_atomic_CNA_profile as get_total_CNA_profile_A_SCM
from Organisms.GA.SCM_Scripts.TC_SCM_Methods import get_CNA_similarity     as get_CNA_similarity_T_SCM
from Organisms.GA.SCM_Scripts.AC_SCM_Methods import get_CNA_similarity     as get_CNA_similarity_A_SCM

class Memory_Operator:

	def __init__(self,memory_operator_information):
		"""

		"""
		if memory_operator_information == {}:
			memory_operator_information['Method'] = 'Off'
		self.method = memory_operator_information['Method']
		if self.method in ['Off','off','OFF','False','false','FALSE',False,None,'None','none','']:
			self.method = 'Off'
		if self.method == 'Off':
			return
		self.rCut = memory_operator_information['rCut']
		self.cut_off_similarity = memory_operator_information['cut_off_similarity']
		if memory_operator_information['SCM Type'] == 'A-SCM':
			self.get_CNA_profile_method = get_total_CNA_profile_A_SCM
			self.get_CNA_similarity_method = get_CNA_similarity_A_SCM
		elif memory_operator_information['SCM Type'] == 'T-SCM':
			self.get_CNA_profile_method = get_total_CNA_profile_T_SCM
			self.get_CNA_similarity_method = get_CNA_similarity_T_SCM
		else:
			exit('Error')

		if self.method == 'after last epoch':
			pass
		elif self.method == 'population':
			self.max_repeat = memory_operator_information['max repeat']
			self.repeats_of_clusters_in_pop_per_generation = {}
		elif self.method == 'clusters in database':
			self.path_to_initial_clusters_database = memory_operator_information['Initial clusters database']
		else:
			exit('memory operator method')
		self.memory_CNA_profiles = []

		self.memory_operator_name = 'Memory_Operator_Data'
		if not os.path.exists(self.memory_operator_name):
			os.mkdir(self.memory_operator_name)
		self.memory_operator_database_name = 'Memory_Operator_Datbase.db'
		self.memory_operator_database_path = self.memory_operator_name+'/'+self.memory_operator_database_name
		self.have_setup_memory_operator = False

	def get_CNA_profile(self,cluster):
		"""

		"""
		return_list = []
		self.get_CNA_profile_method((cluster,self.rCut),return_list)
		cluster_CNA_profile = return_list[0]
		return cluster_CNA_profile

	def setup_from_database(self,last_cluster_name,current_generation):
		"""

		"""
		if self.method == 'Off':
			return
		elif self.method == 'clusters in database':
			#import multiprocessing as mp
			#manager = mp.Manager()
			#return_list = []#manager.list()
			with connect(self.path_to_initial_clusters_database) as database:
				for row in database.select():
					cluster = row.toatoms()
					cluster_CNA_profile = self.get_CNA_profile(cluster)
					self.memory_CNA_profiles.append(cluster_CNA_profile)	
		if not os.path.exists(self.memory_operator_database_path):
			with connect(self.memory_operator_database_path) as database:
				for row in database.select():
					if row.name <= last_cluster_name:
						cluster = row.toatoms()
						cluster_CNA_profile = self.get_CNA_profile(cluster)
						self.memory_CNA_profiles.append(cluster_CNA_profile)	
					else:
						del db[row.id]
		self.have_setup_memory_operator = True

	def __repr__(self):
		return str(self.__dict__)

	def is_too_similar(self,cluster_CNA_profile,memory_CNA_profile):
		"""

		"""
		similarity = self.get_CNA_similarity_method(cluster_CNA_profile,memory_CNA_profile)
		#print('similarity: '+str(similarity))
		return (similarity >= self.cut_off_similarity)

	def check_collection(self,collection):
		"""

		"""
		if (self.method == 'Off') or (len(self.memory_CNA_profiles) == 0):
			return []
		if not self.have_setup_memory_operator:
			exit('Error in memory operaotr')
		offspring_to_remove_because_too_similar_to_a_cluster_in_memory = []
		for index in range(len(collection)):
			offspring = collection[index]
			offspring_CNA_profile = self.get_CNA_profile(offspring)
			for memory_CNA_profile in self.memory_CNA_profiles:
				if self.is_too_similar(offspring_CNA_profile,memory_CNA_profile):
					offspring_to_remove_because_too_similar_to_a_cluster_in_memory.append((offspring.name,index))
					break
		offspring_to_remove_because_too_similar_to_a_cluster_in_memory.sort(key=lambda x:x[1])
		return offspring_to_remove_because_too_similar_to_a_cluster_in_memory

	def remove_similar_clusters_from_offsprng_pool(self,collection,offspring_to_remove_because_too_similar_to_a_cluster_in_memory,print_details):
		"""

		"""
		clusters_being_removed_by_memory_operator = []
		for cluster_name, cluster_index in offspring_to_remove_because_too_similar_to_a_cluster_in_memory[::-1]:
			cluster_being_removed_by_memory_operator = collection.pop(cluster_index)
			cluster_being_removed_by_memory_operator.ever_in_population = False
			cluster_being_removed_by_memory_operator.excluded_because_violates_predation_operator = False
			cluster_being_removed_by_memory_operator.initial_population = False
			cluster_being_removed_by_memory_operator.removed_by_memory_operator = True
			clusters_being_removed_by_memory_operator.append(cluster_being_removed_by_memory_operator)
		if print_details and (not len(clusters_being_removed_by_memory_operator) == 0):
			print('------------------------------------------------------------------------------------------------------------------------------')
			print('The following offspring were removed from the offspring pool because they were too similar to a cluster in the memory database')
			for cluster in clusters_being_removed_by_memory_operator:
				print('Cluster: '+str(cluster.name))
			print('------------------------------------------------------------------------------------------------------------------------------')
		return clusters_being_removed_by_memory_operator

	def is_similar_cluster_in_memory_operator(self,cluster):
		"""

		"""
		if (self.method == 'Off') or (len(self.memory_CNA_profiles) == 0):
			return False
		if not self.have_setup_memory_operator:
			exit('Error in memory operaotr')
		cluster_CNA_profile = self.get_CNA_profile(cluster)
		for memory_CNA_profile in self.memory_CNA_profiles:
			if self.is_too_similar(cluster_CNA_profile,memory_CNA_profile):
				return True
		return False

	def record_clusters_in_population_from_epoch(self, population):
		"""

		"""
		self.add_clusters_to_memory_operator(population)
		'''
		CNA_profiles = []
		for cluster in population:
			cluster_CNA_profile = self.get_CNA_profile(cluster)
			CNA_profiles.append((cluster.name, cluster_CNA_profile))
		similarities = {}
		for index1 in range(len(CNA_profiles)):
			name1, CNA_profile1 = CNA_profiles[index1]
			for index2 in range(index1+1,len(CNA_profiles)):
				name2, CNA_profile2 = CNA_profiles[index1]
				if name1 == name2:
					exit('errpr')
				comparison = (name1,name2) if (name1 < name2) else (name2,name1)
				similarities[comparison] = self.get_CNA_similarity_method(CNA_profile1,CNA_profile2)
		'''

	def look_through_population(self, population):
		"""

		"""
		updated_repeats_of_clusters_in_pop_per_generation = {}
		clusters_to_add_to_memory = []
		for cluster in population:
			number = self.repeats_of_clusters_in_pop_per_generation.get(cluster.name, default=0) + 1
			if number == self.max_repeat:
				clusters_to_add_to_memory.append(cluster)
			else:
				updated_repeats_of_clusters_in_pop_per_generation[cluster.name] = number
		self.repeats_of_clusters_in_pop_per_generation = updated_repeats_of_clusters_in_pop_per_generation
		if not len(clusters_to_add_to_memory):
			self.add_clusters_to_memory_operator(clusters)

	def add_clusters_to_memory_operator(self,clusters):
		"""

		"""
		if self.method == 'Off' or self.method == 'clusters in database':
			return
		with connect(self.memory_operator_database_path) as database:
			for cluster in clusters:
				database.write(cluster,name=cluster.name,gen_made=cluster.gen_made)
				cluster_CNA_profile = self.get_CNA_profile(cluster)
				self.memory_CNA_profiles.append(cluster_CNA_profile)


import os
import subprocess
from shutil import copytree, move

def tail(f, n, offset=0):
	"""
	This method will read the last lines from a text file

	:param f: This is the file path to the path you want to read the end of
	:type  f: str.
	:param n: This is the last lines to read from the document.
	:type  n: int
	:param offset: This is an offset number of line that should at least have been read, but this can be kept as 0 and given in n. For the user to decide. 
	:type  offset: int

	returns lines: These are the line to be read from the end of the document
	rtype   lines: list of str.
	"""
	proc = subprocess.Popen(['tail', '-n', str(n + offset), f], stdout=subprocess.PIPE)
	lines = proc.stdout.readlines()
	return lines

def remove_end_lines_from_text(file, number_of_lines_to_remove):
	"""
	This method removes the last number of lines from the document. 

	For more information, see: https://superuser.com/questions/127786/efficiently-remove-the-last-two-lines-of-an-extremely-large-text-file
	
	:param file: This is the file path to the path you want to read the end of
	:type  file: str.
	:param number_of_lines_to_remove: This is the number of lines to remove from the document.
	:type  number_of_lines_to_remove: int

	"""
	count = 0
	with open(file,'r+b', buffering=0) as f:
		f.seek(0, os.SEEK_END)
		end = f.tell()
		while f.tell() > 0:
			f.seek(-1, os.SEEK_CUR)
			print(f.tell())
			char = f.read(1)
			if char != b'\n' and f.tell() == end:
				print("No change: file does not end with a newline")
				exit(1)
			if char == b'\n':
				count += 1
			if count == number_of_lines_to_remove + 1:
				f.truncate()
				print("Removed " + str(number_of_lines_to_remove) + " lines from end of file")
				return
			f.seek(-1, os.SEEK_CUR)
	if count < number_of_lines_to_remove + 1:
		print("No change: requested removal would leave empty file")
		exit(3)
		
class EnergyProfile:
	"""
	This class is designed to record the energies of the clusters during the genetic algorithm.

	:param collection: This is the specific collection of the genetic algorithm this class will be recording for.
	:type  collection: Organisms.GA.Collection
	:param end_name: Include a suffix to the EnergyProfile.txt filename
	:type  end_name: str.
	"""
	def __init__(self,collection,end_name=None):
		self.path = collection.name
		if end_name == None:
			self.EnergyProfileTXT_path = self.path+"/EnergyProfile.txt"
		else:
			self.EnergyProfileTXT_path = self.path+"/EnergyProfile_"+str(end_name)+".txt"
		self.create()

	def create(self):
		"""
		This definition will create the folders and text files for 
		"""
		if not os.path.exists(self.path):
			os.makedirs(self.path)

	def GA_Starts(self):
		"""
		This will write the energies of the collection into the recording EnergyProfile text file.
		"""
		return

	def add_epoch_note(self):
		"""
		Add a note to the energyprofile to say that an epoch event occurred at this point
		"""
		self.open('a')
		self.energyProfile.write('Restarting due to epoch.\n')
		self.close()

	def add_epoch_note_due_to_population_energy_convergence(self):
		"""
		Add a note to the energyprofile to say that an epoch event occurred at this point due to the energies in the cluster having converged. 
		"""
		self.open('a')
		self.energyProfile.write('Restarting due to epoch. The population has energetically converged.\n')
		self.close()

	def add_found_LES_note(self):
		"""
		Add a note to the energyprofile to say that the LES was located, such that your genetic algorithm will now finish. 
		"""
		self.open('a')
		self.energyProfile.write('Finished prematurely as LES energy found.\n')
		self.close()

	def is_LES_note_in_EnergyProfile(self):
		"""
		Look through the EnergyProfile if it exists for a note saying that the LES had be found.

		:returns True if the EnergyProfile states the LES has been found, otherwise return False.
		:rtype bool.
		"""
		if not os.path.exists(self.EnergyProfileTXT_path):
			return False
		self.open('r')
		for line in self.energyProfile:
			if 'Finished prematurely as LES energy found.' in line:
				return True
		return False

	def add_collection(self,collection,generation_number):
		"""
		This will add information about the collection to the EenrgyProfile.txt

		Inputs:
			collection (Organisms.GA.collection): This is the collection to be added to the EnergyProfile.txt, should be an Offspring_Pool.
			generation_number (int): This is the current generation number of the running Organisms program.
		"""
		self.open('a')
		for cluster in collection:
			self.add_to(cluster,generation_number)
		self.close()

	def add_to(self,cluster,generation_number):
		"""
		This definition will add the information of a cluster. Designed to be used when offspring are created.

		Inputs:
			cluster (Organisms.GA.Cluster): This is the cluster to write information about in the EnergyProfile.txt
			generation_number (int): This is the current generation number of the running Organisms program.
		"""
		self.energyProfile.write("{0}\t{1}\t{2}\n".format(cluster.name, generation_number, cluster.energy))
		
	def open(self,read_type):
		"""
		This will open the EnergyProfile text file that is being recorded to.

		This is designed to be a private definition

		Inputs:
			read_type (str.): This is the way to open the file, either read ('r''), write ('w'), or append ('a').

		"""
		self.energyProfile = open(self.EnergyProfileTXT_path, str(read_type))

	def close(self):
		"""
		This will close the EnergyProfile text file that is being recorded to.

		This is designed to be a private definition
		"""
		self.energyProfile.close()

	def check(self, resume_from_generation, no_offspring_per_generation):
		"""
		Check the EnergyProfile to make sure it is all good to go. As well as if the LES has been located by the genetic algorthm, and to remove any information from the end of the EnergyProfile document that attains to future generations that the algorithm did not complete successfully.

		:param resume_from_generation: The generation that the algorithm is resuming from
		:type  resume_from_generation: int
		:param no_offspring_per_generation: The number of offspring generated per generation. 
		:type  no_offspring_per_generation: int

		"""	
		while True:
			counter = 0
			last_lines_in_EnergyProfile = tail(self.path+'/EnergyProfile.txt',no_offspring_per_generation)[::-1]
			for line in last_lines_in_EnergyProfile:
				if type(line) is bytes:
					line = line.decode()
				if line.startswith('Restarting due to epoch.'):
					counter += 1
					continue
				line = line.split()
				generation = int(line[1])
				if generation > resume_from_generation:
					counter += 1
				elif generation == resume_from_generation:
					break
				elif generation < resume_from_generation:
					print('Error in def check, in class EnergyProfile, in EnergyProfile.py')
					print('The last generation in the '+str(self.EnergyProfileTXT_path)+' is less than the resumed generation.')
					print('Check the EnergyProfile file for this GA run before continuing. There may be an issue here')
					print('Last generation in the EnergyProfile file: '+str(generation))
					print('Generation to resume from: '+str(resume_from_generation))
					import pdb; pdb.set_trace()
					exit('The genetic algorithm will finished without starting.')
			if counter > 0:
				print('Removing the last '+str(counter)+' from '+str(self.path+'/EnergyProfile.txt'))
				remove_end_lines_from_text(self.path+'/EnergyProfile.txt', counter)
			else:
				break
		last_lines_in_EnergyProfile = tail(self.path+'/EnergyProfile.txt',no_offspring_per_generation)
		all_cluster_gen_made = []
		for line in last_lines_in_EnergyProfile:
			line = line.split()
			generation = int(line[1])
			all_cluster_gen_made.append(generation)
		if not (len(set(all_cluster_gen_made)) == 1):
			print('Error in def check, in class EnergyProfile, in EnergyProfile.py')
			print('The last generation in the '+str(self.EnergyProfileTXT_path)+' is not the resumed generation.')
			print('Check the EnergyProfile file for this GA run before continuing. There may be an issue here')
			print('Generations of clusters in proposed population'+str(all_cluster_gen_made))
			print('Generation to resume from: '+str(resume_from_generation))
			import pdb; pdb.set_trace()
			exit('The genetic algorithm will finished without starting.')
		all_cluster_gen_made = all_cluster_gen_made[0]
		if all_cluster_gen_made == resume_from_generation:
			return
		else:
			print('Error in def check, in class EnergyProfile, in EnergyProfile.py')
			print('The last generation in the '+str(self.EnergyProfileTXT_path)+' is less than the resumed generation.')
			print('Check the EnergyProfile file for this GA run before continuing. There may be an issue here')
			print('Last generation in the EnergyProfile file: '+str(all_cluster_gen_made))
			print('Generation to resume from: '+str(resume_from_generation))
			import pdb; pdb.set_trace()
			exit('The genetic algorithm will finished without starting.')

	def get_current_generation_and_last_cluster_generated_from_EnergyProfile(self):
		"""
		Get the number of the last generation and the name of the last cluster that was recorded in the EnergyProfile

		:returns The last recorded generation, and the name of the last recorded cluster.
		:rtype   int, int

		"""
		last_lines_in_EnergyProfile = tail(self.path+'/EnergyProfile.txt',1)
		#import pdb; pdb.set_trace()
		#line = last_lines_in_EnergyProfile.readline()
		#line = line.rstrip().split()
		line = last_lines_in_EnergyProfile[0].rstrip().split()
		name       = int(line[0])
		generation = int(line[1])
		return generation, name



import sys

def get_predation_operator(predation_information, fitness_information, population, no_of_cpus, print_details):
	"""
	This def will set up the predation operator to be by the genetic algorithm. This is dependent on the entry for the 'Predation_Switch' 
	given in the predation_information. 

	:param predation_information: All the informatino about the predation operator.
	:type  predation_information: dict.
	:param fitness_information: All the information about the fitness operator
	:type  fitness_information: dict.
	:param population: The population
	:type  population: Organisms.GA.Population
	:param no_of_cpus: The number of cpus that are to be used by your genetic algorithm run.
	:type  no_of_cpus: int
	:param print_details: Print information to the terminal
	:type  print_details: bool.

	:returns: The predation operator object
	:rtype:    Organisms.GA.Predation_Operators.Predation_Operator

	"""
	predation_switch = predation_information['Predation Operator']
	if predation_switch == 'Off':
		from Organisms.GA.Predation_Operators.No_Predation_Operator import No_Predation_Operator
		return No_Predation_Operator(predation_information, population, print_details)
	elif predation_switch == 'Energy':
		from Organisms.GA.Predation_Operators.Energy_Predation_Operator import Energy_Predation_Operator
		return Energy_Predation_Operator(predation_information, population, print_details)
	if predation_switch == 'IDCM':
		from Organisms.GA.Predation_Operators.IDCM_Predation_Operator import IDCM_Predation_Operator
		return IDCM_Predation_Operator(predation_information, population, no_of_cpus, print_details)
	elif predation_switch == 'SCM':
		from Organisms.GA.Predation_Operators.SCM_Predation_Operator import SCM_Predation_Operator
		return SCM_Predation_Operator(predation_information, fitness_information, population, no_of_cpus, print_details)#,no_of_generations)
	else:
		error_string =  'Error in choosing Predation Operators.\n'
		error_string += 'Predation_Operator must be either:'
		error_string += '\t"Off" - No Predation Operator is used.\n'
		error_string += '\t"Energy" - Based on no two clusters can have the same energy in the population.\n'
		error_string += '\t"IDCM" - Based on no two clusters can be structurally identical in the population based on the comparison of those two structuress EDM\'. See Manual for more information.\n'
		error_string += '\t"SCM" - No two structures can be Geometrically Similar based on the Structral Comarison Method. See Manual for more information.\n'
		error_string += 'Check this.\n'
		error_string += 'Predation_Operator = ' + str(predation_switch)+'\n'
		#import pdb, traceback, sys
		#extype, value, tb = sys.exc_info()
		#traceback.print_exc()
		#pdb.post_mortem(tb)
		print(error_string,file=sys.stderr)
		import pdb; pdb.set_trace()
		exit()

def get_fitness_operator(fitness_information, predation_operator, population, generations, no_of_cpus, print_details):
	"""
	This def will set up the fitness operator to be by the genetic algorithm. This is dependent on the entry for the 'Predation_Switch' 
	given in the fitness_information. 

	:param fitness_information: All the information about the fitness operator
	:type  fitness_information: dict.
	:param predation_information: All the informatino about the predation operator.
	:type  predation_information: dict.
	:param population: The population
	:type  population: Organisms.GA.Population
	:param generations: The number of generations being run
	:param generations: int
	:param no_of_cpus: The number of cpus that are to be used by your genetic algorithm run.
	:type  no_of_cpus: int
	:param print_details: Print information to the terminal
	:type  print_details: bool.

	:returns: The fitness operator object
	:rtype:   Organisms.GA.Fitness_Operators.Fitness_Operator

	"""
	fitness_switch = fitness_information['Fitness Operator']
	if fitness_switch == 'Energy':
		from Organisms.GA.Fitness_Operators.Energy_Fitness_Operator import Energy_Fitness_Operator
		return Energy_Fitness_Operator(fitness_information, predation_operator, population, print_details)
	elif fitness_switch == 'SCM + Energy':
		from Organisms.GA.Fitness_Operators.SCM_and_Energy_Fitness_Operator import SCM_and_Energy_Fitness_Operator
		return SCM_and_Energy_Fitness_Operator(fitness_information, predation_operator, population, generations, no_of_cpus, print_details)
	else:
		error_string =  'Error in choosing Fitness Operators.\n'
		error_string += 'Predation_Operator must be either:\n'
		error_string += '\t"Energy" - The fitness of a cluster is based on its energy only.\n'
		error_string += '\t"SCM + Energy" - The fitness of a cluster is based on its energy and structural diversity as determined by the SCM.\n'
		error_string += 'Check this.\n'
		error_string += 'Fitness Operator = ' + str(fitness_switch)+'\n'
		#import pdb, traceback, sys
		#extype, value, tb = sys.exc_info()
		#traceback.print_exc()
		#pdb.post_mortem(tb)
		print(error_string,file=sys.stderr)
		import pdb; pdb.set_trace()
		exit()

def get_predation_and_fitness_operators(predation_information, fitness_information, population, generations, no_of_cpus, print_details):
	"""
	This def will set up the predation and fitness operators to be by the genetic algorithm. 

	:param predation_information: All the informatino about the predation operator.
	:type  predation_information: dict.
	:param fitness_information: All the information about the fitness operator
	:type  fitness_information: dict.
	:param population: The population
	:type  population: Organisms.GA.Population
	:param generations: The number of generations being run
	:param generations: int
	:param no_of_cpus: The number of cpus that are to be used by your genetic algorithm run.
	:type  no_of_cpus: int
	:param print_details: Print information to the terminal
	:type  print_details: bool.

	:returns: The predation operator object and the fitness operator object
	:rtype:    A Organisms.GA.Predation_Operators.Predation_Operator object and a Organisms.GA.Fitness_Operators.Fitness_Operator object

	"""
	if not 'Use Predation Information' in fitness_information:
		fitness_information['Use Predation Information'] = False
	predation_operator = get_predation_operator(predation_information, fitness_information, population, no_of_cpus, print_details)
	fitness_operator = get_fitness_operator(fitness_information, predation_operator, population, generations, no_of_cpus, print_details)
	return predation_operator, fitness_operator

